import os
import numpy as np
import cv2
from PIL import Image
from torch.utils.data import Dataset
from ..utils import LOG

class FileLineDataset(Dataset):
    def __init__(self, deepvac_config):
        self.path_prefix = deepvac_config.fileline_data_path_prefix
        self.fileline_path = deepvac_config.fileline_path
        self.transform = deepvac_config.transform
        self.samples = []
        mark = []

        with open(self.fileline_path) as f:
            for line in f:
                label = self._buildLabelFromLine(line)
                self.samples.append(label)
                mark.append(label[1])

        self.len = len(self.samples)
        self.class_num = len(np.unique(mark))
        LOG.logI('FileLineDataset size: {} / {}'.format(self.len, self.class_num))

    def _buildLabelFromLine(self, line):
        line = line.strip().split(" ")
        return [line[0], int(line[1])]

    def __getitem__(self, index):
        path, target = self.samples[index]
        abs_path = os.path.join(self.path_prefix, path)
        return self._buildSampleFromPath(abs_path), target

    def _buildSampleFromPath(self, abs_path):
        #we just set default loader with Pillow Image
        sample = Image.open(abs_path).convert('RGB')
        if self.transform is not None:
            sample = self.transform(sample)
        return sample

    def __len__(self):
        return self.len

class FileLineCvStrDataset(FileLineDataset):
    def _buildLabelFromLine(self, line):
        line = line.strip().split(" ", 1)
        return [line[0], line[1]]

    def _buildSampleFromPath(self, abs_path):
        #we just set default loader with Pillow Image
        sample = cv2.imread(abs_path)
        if self.transform is not None:
            sample = self.transform(sample)
        return sample


import numpy as np
import torch
import torch.nn as nn
from ..utils import LOG
from .weights_init import initWeightsKaiming
from .conv_layer import Conv2dBNWithName

class RepVGGModelConvert(object):
    def __init__(self):
        self.all_weights = {}

    def _isValidName(self, name):
        if name[0] == '.':
            return False
        if '.rbr_dense' in name:
            return False
        if '.rbr_1x1' in name:
            return False
        if '.rbr_identity' in name:
            return False
        return True

    def _addReparamPart(self, name, module):
        kernel, bias = module.repvgg_convert()
        self.all_weights[name + '.rbr_reparam.weight'] = kernel
        self. all_weights[name + '.rbr_reparam.bias'] = bias

    def _addOtherPart(self, name, module):
        for p_name, p_tensor in module.named_parameters():
            full_name = name + '.' + p_name
            if self._isValidName(full_name) and full_name not in self.all_weights:
                self.all_weights[full_name] = p_tensor.detach()
        for p_name, p_tensor in module.named_buffers():
            full_name = name + '.' + p_name
            if self._isValidName(full_name) and full_name not in self.all_weights:
                self.all_weights[full_name] = p_tensor

    def __call__(self, train_model:torch.nn.Module, deploy_model:torch.nn.Module, save_path=None):
        self.all_weights = {}
        for name, module in train_model.named_modules():
            if hasattr(module, 'repvgg_convert'):
                self._addReparamPart(name, module)
            else:
                self._addOtherPart(name, module)

        deploy_model.load_state_dict(self.all_weights)
        if save_path is not None:
            torch.save(deploy_model.state_dict(), save_path)

        return deploy_model

class RepVGGBlock(nn.Module):
    def __init__(self, in_channels, out_channels, stride=1, dilation=1, groups=1, padding_mode='zeros', deploy=False):
        super(RepVGGBlock, self).__init__()
        self.deploy = deploy
        self.groups = groups
        self.in_channels = in_channels

        kernel_size = 3
        padding = 1

        padding_11 = padding - kernel_size // 2

        self.relu = nn.ReLU()

        if deploy:
            self.rbr_reparam = nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size, stride=stride,
                padding=padding, dilation=dilation, groups=groups, bias=True, padding_mode=padding_mode)

        else:
            self.rbr_identity = nn.BatchNorm2d(num_features=in_channels) if out_channels == in_channels and stride == 1 else None
            self.rbr_dense = Conv2dBNWithName(in_channels, out_channels, kernel_size, stride, padding, groups)
            self.rbr_1x1 = Conv2dBNWithName(in_channels, out_channels, 1, stride, padding_11, groups)

    def forward(self, inputs):
        if self.deploy:
            return self.relu(self.rbr_reparam(inputs))

        id_out = self.rbr_identity(inputs) if self.rbr_identity else 0

        return self.relu(self.rbr_dense(inputs) + self.rbr_1x1(inputs) + id_out)

#   This func derives the equivalent kernel and bias in a DIFFERENTIABLE way.
#   You can get the equivalent kernel and bias at any time and do whatever you want,
    #   for example, apply some penalties or constraints during training, just like you do to the other models.
#   May be useful for quantization or pruning.
    def get_equivalent_kernel_bias(self):
        kernel3x3, bias3x3 = self._fuse_bn_tensor(self.rbr_dense)
        kernel1x1, bias1x1 = self._fuse_bn_tensor(self.rbr_1x1)
        kernelid, biasid = self._fuse_bn_tensor(self.rbr_identity)
        return kernel3x3 + self._pad_1x1_to_3x3_tensor(kernel1x1) + kernelid, bias3x3 + bias1x1 + biasid

    def _pad_1x1_to_3x3_tensor(self, kernel_1x1):
        if kernel_1x1 is None:
            return 0
        return torch.nn.functional.pad(kernel_1x1, [1,1,1,1])

    def _fuse_bn_tensor_for_sequential(self, branch):
        kernel = branch.conv.weight
        running_mean = branch.bn.running_mean
        running_var = branch.bn.running_var
        gamma = branch.bn.weight
        beta = branch.bn.bias
        eps = branch.bn.eps

        std = (running_var + eps).sqrt()
        t = (gamma / std).reshape(-1, 1, 1, 1)
        return kernel * t, beta - running_mean * gamma / std

    def _fuse_bn_tensor_for_bn(self, branch):
        if not hasattr(self, 'id_tensor'):
            input_dim = self.in_channels // self.groups
            kernel_value = np.zeros((self.in_channels, input_dim, 3, 3), dtype=np.float32)
            for i in range(self.in_channels):
                kernel_value[i, i % input_dim, 1, 1] = 1
            self.id_tensor = torch.from_numpy(kernel_value).to(branch.weight.device)
        kernel = self.id_tensor
        running_mean = branch.running_mean
        running_var = branch.running_var
        gamma = branch.weight
        beta = branch.bias
        eps = branch.eps

        std = (running_var + eps).sqrt()
        t = (gamma / std).reshape(-1, 1, 1, 1)
        return kernel * t, beta - running_mean * gamma / std

    def _fuse_bn_tensor(self, branch):
        if branch is None:
            return 0, 0
        
        if isinstance(branch, nn.Sequential):
            return self._fuse_bn_tensor_for_sequential(branch)
        elif isinstance(branch, nn.BatchNorm2d):
            return self._fuse_bn_tensor_for_bn(branch)

    def repvgg_convert(self):
        kernel, bias = self.get_equivalent_kernel_bias()
        return kernel.detach(), bias.detach()

class RepVGG(nn.Module):
    def __init__(self, class_num=1000, deploy=False):
        super(RepVGG, self).__init__()

        self.deploy = deploy
        self.auditConfig()

        self.in_planes = min(64, self.cfgs[0][0])
        self.stage0 = RepVGGBlock(in_channels=3, out_channels=self.in_planes, stride=2, deploy=self.deploy)
        self.cur_layer_idx = 1

        layers = []
        for planes, num_blocks, stride in self.cfgs:
            layers.extend(self._make_layer(planes, num_blocks, stride))
        
        self.layer = nn.Sequential(*layers)

        self.avgpool = nn.AdaptiveAvgPool2d(output_size=1)
        self.linear = nn.Linear(self.cfgs[-1][0], class_num)

        initWeightsKaiming(self)

    def auditConfig(self):
        self.cfgs = None
        self.override_groups_map = None
        LOG.logE("You must reimplement auditConfig() to initialize self.cfgs and self.override_groups_map", exit=True)

    def _make_layer(self, planes, num_blocks, stride):
        strides = [stride] + [1]*(num_blocks-1)
        blocks = []
        for stride in strides:
            cur_groups = self.override_groups_map.get(self.cur_layer_idx, 1)
            blocks.append(RepVGGBlock(in_channels=self.in_planes, out_channels=planes, stride=stride, groups=cur_groups, deploy=self.deploy))
            self.in_planes = planes
            self.cur_layer_idx += 1
        return blocks

    def forward(self, x):
        out = self.stage0(x)
        out = self.layer(out)
        out = self.avgpool(out)
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        return out

class RepVGGASmall(RepVGG):
    def __init__(self, class_num=1000, deploy=False):
        super(RepVGGASmall, self).__init__(class_num, deploy)

    def auditConfig(self):
        self.cfgs = [
            # planes, num_blocks, stride
            [48, 2, 2],
            [96, 4, 2],
            [192, 14, 2],
            [1280, 1, 2]
        ]
        self.override_groups_map = dict()

class RepVGGAMedium(RepVGG):
    def __init__(self, class_num=1000, deploy=False):
        super(RepVGGAMedium, self).__init__(class_num, deploy)
    
    def auditConfig(self):
        self.cfgs = [
            # planes, num_blocks, stride
            [64, 2, 2],
            [128, 4, 2],
            [256, 14, 2],
            [1280, 1, 2]
        ]
        self.override_groups_map = dict()

class RepVGGALarge(RepVGG):
    def __init__(self, class_num=1000, deploy=False):
        super(RepVGGALarge, self).__init__(class_num, deploy)

    def auditConfig(self):
        self.cfgs = [
            # planes, num_blocks, stride
            [96, 2, 2],
            [192, 4, 2],
            [384, 14, 2],
            [1408, 1, 2]
        ]
        self.override_groups_map = dict()

class RepVGGBSmall(RepVGG):
    def __init__(self, class_num=1000, deploy=False):
        super(RepVGGBSmall, self).__init__(class_num, deploy)

    def auditConfig(self):
        self.cfgs = [
            # planes, num_blocks, stride
            [64, 4, 2],
            [128, 6, 2],
            [256, 16, 2],
            [1280, 1, 2]
        ]
        self.override_groups_map = dict()

class RepVGGBMedium(RepVGG):
    def __init__(self, class_num=1000, deploy=False):
        super(RepVGGBMedium, self).__init__(class_num, deploy)

    def auditConfig(self):
        self.cfgs = [
            # planes, num_blocks, stride
            [128, 4, 2],
            [256, 6, 2],
            [512, 16, 2],
            [2048, 1, 2]
        ]
        self.override_groups_map = dict()

class RepVGGBLarge(RepVGG):
    def __init__(self, class_num=1000, deploy=False):
        super(RepVGGBLarge, self).__init__(class_num, deploy)

    def auditConfig(self):
        self.cfgs = [
            # planes, num_blocks, stride
            [160, 4, 2],
            [320, 6, 2],
            [640, 16, 2],
            [2560, 1, 2]
        ]
        self.override_groups_map = dict()

class RepVGGBMediumG2(RepVGG):
    def __init__(self, class_num=1000, deploy=False):
        super(RepVGGBMediumG2, self).__init__(class_num, deploy)

    def auditConfig(self):
        self.cfgs = [
            # planes, num_blocks, stride
            [128, 4, 2],
            [256, 6, 2],
            [512, 16, 2],
            [2048, 1, 2]
        ]
        optional_groupwise_layers = [2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26]
        self.override_groups_map = {l: 2 for l in optional_groupwise_layers}

class RepVGGBMediumG4(RepVGG):
    def __init__(self, class_num=1000, deploy=False):
        super(RepVGGBMediumG4, self).__init__(class_num, deploy)

    def auditConfig(self):
        self.cfgs = [
            # planes, num_blocks, stride
            [128, 4, 2],
            [256, 6, 2],
            [512, 16, 2],
            [2048, 1, 2]
        ]
        optional_groupwise_layers = [2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26]
        self.override_groups_map = {l: 4 for l in optional_groupwise_layers}

class RepVGGBLargeG2(RepVGG):
    def __init__(self, class_num=1000, deploy=False):
        super(RepVGGBLargeG2, self).__init__(class_num, deploy)

    def auditConfig(self):
        self.cfgs = [
            # planes, num_blocks, stride
            [160, 4, 2],
            [320, 6, 2],
            [640, 16, 2],
            [2560, 1, 2]
        ]
        optional_groupwise_layers = [2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26]
        self.override_groups_map = {l: 2 for l in optional_groupwise_layers}

class RepVGGBLargeG4(RepVGG):
    def __init__(self, class_num=1000, deploy=False):
        super(RepVGGBLargeG4, self).__init__(class_num, deploy)

    def auditConfig(self):
        self.cfgs = [
            # planes, num_blocks, stride
            [160, 4, 2],
            [320, 6, 2],
            [640, 16, 2],
            [2560, 1, 2]
        ]
        optional_groupwise_layers = [2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26]
        self.override_groups_map = {l: 4 for l in optional_groupwise_layers}

if __name__ == '__main__':
    model = RepVGGASmall()
    LOG.logI(model)
    x = torch.randn((1,3,32,320))
    LOG.logI(model(x).shape)



import os
import cv2
from torch.utils.data import Dataset

class OsWalkDataset(Dataset):
    def __init__(self, deepvac_config):
        super(OsWalkDataset, self).__init__()
        self.transform = deepvac_config.transform
        self.files = []
        for subdir, dirs, fns in os.walk(deepvac_config.input_dir):
            for fn in fns:
                self.files.append(os.path.join(subdir, fn))

    def __len__(self):
        return len(self.files)

    def __getitem__(self, index):
        filepath = self.files[index]
        sample = cv2.imread(filepath)
        if self.transform is not None:
            sample = self.transform(sample)
        return sample


from ..utils import is_ddp

if is_ddp:
    from .deepvac import DeepvacDDP as DeepvacTrain, Deepvac
else:
    from .deepvac import DeepvacTrain, Deepvac

from .config import config, AttrDict
from .report import *


import numpy as np
import torch
import torch.nn as nn
from .weights_init import initWeightsKaiming
#function begin

def adjust_block_compatibility(width_per_stage, bot_muls_per_stage, group_w_per_stage):
    """Adjusts the compatibility of widths, bottlenecks, and groups."""
    assert len(width_per_stage) == len(bot_muls_per_stage) == len(group_w_per_stage)
    assert all(w > 0 and b > 0 and g > 0 for w, b, g in zip(width_per_stage, bot_muls_per_stage, group_w_per_stage))
    assert all(b < 1 or b % 1 == 0 for b in bot_muls_per_stage)
    vs = [int(max(1, w * b)) for w, b in zip(width_per_stage, bot_muls_per_stage)]
    group_w_per_stage = [int(min(g, v)) for g, v in zip(group_w_per_stage, vs)]
    ms = [np.lcm(g, int(b)) if b > 1 else g for g, b in zip(group_w_per_stage, bot_muls_per_stage)]
    vs = [max(m, int(round(v / m) * m)) for v, m in zip(vs, ms)]
    width_per_stage = [int(v / b) for v, b in zip(vs, bot_muls_per_stage)]
    assert all(w * b % g == 0 for w, b, g in zip(width_per_stage, bot_muls_per_stage, group_w_per_stage))
    return width_per_stage, bot_muls_per_stage, group_w_per_stage

def generate_regnet(anynet_slope, anynet_initial_width, channel_controller, depth, divisor=8):
    """Generates per stage widths and depths from RegNet parameters."""
    assert anynet_slope >= 0 and anynet_initial_width > 0 and channel_controller > 1 and anynet_initial_width % divisor == 0
    # Generate continuous per-block ws
    ws_cont = np.arange(depth) * anynet_slope + anynet_initial_width
    # Generate quantized per-block ws
    ks = np.round(np.log(ws_cont / anynet_initial_width) / np.log(channel_controller))
    ws_all = anynet_initial_width * np.power(channel_controller, ks)
    ws_all = np.round(np.divide(ws_all, divisor)).astype(int) * divisor
    # Generate per stage ws and ds (assumes ws_all are sorted)
    ws, ds = np.unique(ws_all, return_counts=True)
    # Compute number of actual stages and total possible stages
    num_stages, total_stages = len(ws), ks.max() + 1
    # Convert numpy arrays to lists and return
    ws, ds, ws_all, ws_cont = (x.tolist() for x in (ws, ds, ws_all, ws_cont))
    return ws, ds, num_stages, total_stages, ws_all, ws_cont

class AnyHead(nn.Module):
    """AnyNet head: AvgPool, 1x1."""

    def __init__(self, in_planes, class_num, head_w):
        super(AnyHead, self).__init__()
        self.head_w = head_w
        if head_w>0:
            self.conv = nn.Conv2d(in_planes, head_w, 1, stride=1, padding=0, groups=1, bias=False)
            self.bn = nn.BatchNorm2d(head_w, eps=1e-5, momentum=0.1)
            self.relu = nn.ReLU(inplace=True)
            in_planes = head_w
        self.avg_pool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(in_planes, class_num, bias=True)

    def forward(self, x):
        x = self.relu(self.bn(self.conv(x))) if self.head_w > 0 else x
        x = self.avg_pool(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)
        return x

class SE(nn.Module):
    """Squeeze-and-Excitation (SE) block: AvgPool, FC, ReLU, FC, Sigmoid."""

    def __init__(self, in_planes, se_planes):
        super(SE, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d((1, 1))
        self.f_ex = nn.Sequential(
            nn.Conv2d(in_planes, se_planes, 1, bias=True),
            nn.ReLU(inplace=True),
            nn.Conv2d(se_planes, in_planes, 1, bias=True),
            nn.Sigmoid(),
        )

    def forward(self, x):
        return x * self.f_ex(self.avg_pool(x))

class BottleneckTransform(nn.Module):
    """Bottleneck transformation: 1x1, 3x3 [+SE], 1x1."""

    def __init__(self, in_planes, out_planes, stride, bottom_mul, group_width, se_ratio):
        super(BottleneckTransform, self).__init__()
        w_b = int(round(out_planes * bottom_mul))
        g = w_b // group_width
        self.a = nn.Conv2d(in_planes, w_b, 1, stride=1, padding=0, bias=False)
        self.a_bn = nn.BatchNorm2d(w_b, eps=1e-5, momentum=0.1)
        self.a_relu = nn.ReLU(inplace=True)
        self.b = nn.Conv2d(w_b, w_b, 3, stride=stride, padding=1, groups=g, bias=False)
        self.b_bn = nn.BatchNorm2d(w_b, eps=1e-5, momentum=0.1)
        self.b_relu = nn.ReLU(inplace=True)
        if se_ratio:
            se_planes = int(round(in_planes * se_ratio))
            self.se = SE(w_b, se_planes)
        self.c = nn.Conv2d(w_b, out_planes, 1, stride=1, padding=0, bias=False)
        self.c_bn = nn.BatchNorm2d(out_planes, eps=1e-5, momentum=0.1)
        self.c_bn.final_bn = True

    def forward(self, x):
        for layer in self.children():
            x = layer(x)
        return x

class ResBottleneckBlock(nn.Module):
    """Residual bottleneck block: x + F(x), F = bottleneck transform."""

    def __init__(self, in_planes, out_planes, stride, bottom_mul=1.0, group_width=1, se_ratio=None):
        super(ResBottleneckBlock, self).__init__()
        # Use skip connection with projection if shape changes
        self.proj_block = (in_planes != out_planes) or (stride != 1)
        if self.proj_block:
            self.proj = nn.Conv2d(in_planes, out_planes, 1, stride=stride, padding=0, bias=False)
            self.bn = nn.BatchNorm2d(out_planes, eps=1e-5, momentum=0.1)
        self.f = BottleneckTransform(in_planes, out_planes, stride, bottom_mul, group_width, se_ratio)
        self.relu = nn.ReLU(True)

    def forward(self, x):
        if self.proj_block:
            x = self.bn(self.proj(x)) + self.f(x)
        else:
            x = x + self.f(x)
        x = self.relu(x)
        return x

class SimpleStemIN(nn.Module):
    """Simple stem for ImageNet: 3x3, BN, ReLU."""

    def __init__(self, in_planes, out_planes):
        super(SimpleStemIN, self).__init__()
        self.conv = nn.Conv2d(in_planes, out_planes, 3, stride=2, padding=1, bias=False)
        self.bn = nn.BatchNorm2d(out_planes, eps=1e-5, momentum=0.1)
        self.relu = nn.ReLU(True)

    def forward(self, x):
        for layer in self.children():
            x = layer(x)
        return x

class AnyStage(nn.Module):
    """AnyNet stage (sequence of blocks w/ the same output shape)."""

    def __init__(self, in_planes, out_planes, stride, depth, block_fun, bottom_mul, group_width, se_ratio):
        super(AnyStage, self).__init__()
        for i in range(depth):
            b_stride = stride if i == 0 else 1
            b_w_in = in_planes if i == 0 else out_planes
            name = "b{}".format(i + 1)
            self.add_module(name, block_fun(b_w_in, out_planes, b_stride, bottom_mul, group_width, se_ratio))

    def forward(self, x):
        for block in self.children():
            x = block(x)
        return x

class RegNet(nn.Module):
    def __init__(self, anynet_slope, anynet_initial_width, channel_controller, depth, group_w, class_num=0, head_w=0):
        super(RegNet, self).__init__()
        stem_w = 32
        se_ratio = 0.25
        width_per_stage, depth_per_stage = generate_regnet(anynet_slope, anynet_initial_width, channel_controller, depth)[0:2]
        group_w_per_stage = [group_w for _ in width_per_stage]
        bot_muls_per_stage = [1.0 for _ in width_per_stage]
        stride_per_stage = self.auditConfig(width_per_stage)
        width_per_stage, bot_muls_per_stage, group_w_per_stage = adjust_block_compatibility(width_per_stage, bot_muls_per_stage, group_w_per_stage)
        stage_params = list(zip(depth_per_stage, width_per_stage, stride_per_stage, bot_muls_per_stage, group_w_per_stage))

        self.stem = SimpleStemIN(3, stem_w)
        prev_w = stem_w
        for i, (depth, w, s, bottom_mul, group_width) in enumerate(stage_params):
            name = "s{}".format(i + 1)
            self.add_module(name, AnyStage(prev_w, w, s, depth, ResBottleneckBlock, bottom_mul, group_width, se_ratio))
            prev_w = w

        if class_num > 0:
            self.head = AnyHead(in_planes=prev_w, class_num=class_num, head_w=head_w)

        initWeightsKaiming(self)
    
    def auditConfig(self, width_per_stage):
        stride_per_stage = [2 for _ in width_per_stage]
        return stride_per_stage

    def forward(self, x):
        for module in self.children():
            x = module(x)
        return x

class RegNetSmall(RegNet):
    def __init__(self, class_num=0, head_w=0):
        super(RegNetSmall, self).__init__(27.89, 48, 2.09, 16, 8, class_num, head_w)

class RegNetMedium(RegNet):
    def __init__(self, class_num=0, head_w=0):
        super(RegNetMedium, self).__init__(20.71, 48, 2.65, 27, 24, class_num, head_w)

class RegNetLarge(RegNet):
    def __init__(self, class_num=0, head_w=0):
        super(RegNetLarge, self).__init__(31.41, 96, 2.24, 22, 64, class_num, head_w)

class RegNetOCR(RegNet):
    def __init__(self, class_num=0, head_w=0):
        super(RegNetOCR, self).__init__(27.89, 48, 2.09, 16, 8, class_num, head_w)

    def auditConfig(self, width_per_stage):
        stride_per_stage = [(2,1),(2,1),(2,1),2]
        return stride_per_stage

if __name__ == '__main__':
    model = RegNetOCR()
    LOG.logI(model)
    x = torch.randn((1,3,32,320))
    LOG.logI(model(x).shape)



import torch.nn as nn

def initWeightsKaiming(civilnet):
    for m in civilnet.modules():
        if isinstance(m, nn.Conv2d):
            nn.init.kaiming_normal_(m.weight, mode='fan_out')
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        elif isinstance(m, nn.BatchNorm2d):
            if m.affine:
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
        elif isinstance(m, nn.Linear):
            nn.init.normal_(m.weight, 0, 0.01)
            if m.bias is not None:
                nn.init.zeros_(m.bias)

def initWeightsNormal(civilnet):
    for m in civilnet.modules():
        if isinstance(m, nn.Conv2d):
            nn.init.normal_(m.weight, 0, 0.02)
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        elif isinstance(m, nn.BatchNorm2d):
            if m.affine:
                nn.init.normal_(m.weight, 1.0, 0.02)
                nn.init.zeros_(m.bias)


import time
import warnings
import numpy as np
import scipy.sparse as ss
from collections import OrderedDict
from ..utils import LOG

class Report(object):
    def __init__(self, ds_name = "Unknown", total_num = 0):
        self.total_num = total_num
        self.ds_name = ds_name
        self.initNumeratorKeys()
        self.initFuncDict()
        self.initReportFormat()
        self.reset()

    def add(self, gt, predict):
        """
        第一个参数gt为groundtruth 的label；如果要测试的ds图片并没有注册db，则gt为None；
            否则gt为对应的str(e.g. "famous_000001");
        第二个参数predict为预测出来的label；如果要测试的ds图片没有匹配到任何db，则predict为None；
            否则predict为对应的str(e.g. "famous_000001");
        """
        self.process_num += 1

        k = (bool(gt), bool(predict))
        self.func_dict[k](gt, predict)

        assert sum(self.metrics.values()) == self.process_num, "internal error"
        assert self.process_num <= self.total_num, "total num error"
        return self

    def __call__(self):
        self.report_dict['duration'] = "%.3f"%(time.time() - self.start_time)

        f = lambda l: sum([org for gemfield, org in self.metrics.items() if gemfield in l])
        ff = lambda k: (f(self.numerator_keys[k]) / f(self.denominator_keys[k]) ) if f(self.denominator_keys[k]) != 0 else 0 

        report_keys = [k for k,v in self.report_dict.items()]
        report_values = [ff(k) if not v else v for k,v in self.report_dict.items() ]

        LOG.logI(self.report_str.format(*report_keys, *report_values))

        assert self.process_num == self.total_num, "total num error"

    def initReportFormat(self):
        self.report_str = '''
|{}|{}|{}|{}|{}|{}|{}|{}|
|--|--|--|--|--|--|--|--|
|{}|{}|{}|{}|{}|{}|{}|{}|
        '''
    def initMetricsDict(self):
        self.metric_keys = ['tp','tn','fp','fn']
        self.metrics = dict.fromkeys(self.metric_keys)
        for k in self.metrics:
            self.metrics[k] = 0

    def reset(self):
        self.process_num = 0
        self.start_time = time.time()
        self.initMetricsDict()
        self.initReportDict()

    def initNumeratorKeys(self):
        self.denominator_keys = dict()
        self.numerator_keys = dict()
        self.denominator_keys['accuracy'] = self.denominator_keys['miss'] = self.denominator_keys['error'] = ['tp','fp','tn','fn']
        self.denominator_keys['precision'] = ['tp','fp']
        self.denominator_keys['recall'] =  ['tp','fn']
        self.numerator_keys['accuracy'] = ['tp','tn']
        self.numerator_keys['precision'] = self.numerator_keys['recall'] = ['tp']
        self.numerator_keys['miss'] = ['fn']
        self.numerator_keys['error'] = ['fp','fn']

    def initReportDict(self):
        self.report_dict = OrderedDict()
        self.report_dict['dataset'] = self.ds_name
        self.report_dict['total'] = self.total_num
        self.report_dict['duration'] = 0
        self.report_dict['accuracy'] = 0
        self.report_dict['precision'] = 0
        self.report_dict['recall'] = 0
        self.report_dict['miss'] = 0
        self.report_dict['error'] = 0

    def initFuncDict(self):
        self.func_dict = dict()
        #预测结果大于阈值，且正确的是TP，且错误的是FP；
        self.func_dict[(True,True)] = lambda gt,predict : self.metrics.update({'tp':self.metrics['tp'] + 1}) if gt == predict else self.metrics.update({'fp':self.metrics['fp'] + 1})
        #FP,预测结果大于阈值，且错误；
        self.func_dict[(False,True)] = lambda gt,predict : self.metrics.update({'fp':self.metrics['fp'] + 1})
        #TN,预测结果小于阈值，且正确；也即没注册db的ds 没有匹配 任何db；
        self.func_dict[(False,False)] = lambda gt,predict : self.metrics.update({'tn':self.metrics['tn'] + 1})
        #FN,预测结果小于阈值，且错误；也即已注册db的ds 没有匹配 任何db；
        self.func_dict[(True,False)] = lambda gt,predict : self.metrics.update({'fn':self.metrics['fn'] + 1})

class FaceReport(Report):
    def __init__(self, ds_name = "Unknown", total_num = 0):
        super(FaceReport,self).__init__(ds_name, total_num)

#For SYSZUXocr test dataset.
class OcrReport(Report):
    def __init__(self, ds_name = "Unknown", total_num = 0):
        super(OcrReport,self).__init__(ds_name, total_num)

    def reset(self):
        super(OcrReport, self).reset()

    def add(self, gt, predict):
        #per whole sequence
        if not isinstance(gt, list):
            gt = [gt]
        if not isinstance(predict, list):
            predict = [predict]
        assert len(gt) == len(predict), 'len(gt)={} not equal to len(predict)={}'.format(len(gt),len(predict))
        for i in range(len(gt)):
            super(OcrReport, self).add(gt[i], predict[i])
            #per character
            gt_len = len(gt[i])
            self.char_report_dict['total_per_char']  += gt_len

            edit_distance = self.levenshteinDistance(gt[i], predict[i])
            correct_len =  gt_len - edit_distance
            if correct_len > 0:
                self.char_report_dict['correct_per_char'] += correct_len

    def initReportFormat(self):
        super(OcrReport, self).initReportFormat()
        self.char_report_str = '''
|{}|{}|{}|{}|
|--|--|--|--|
|{}|{}|{}|{}|
        '''

    def initReportDict(self):
        super(OcrReport, self).initReportDict()
        self.char_report_dict = OrderedDict()
        self.char_report_dict['dataset'] = self.ds_name
        self.char_report_dict['total_per_char'] = 0
        self.char_report_dict['correct_per_char'] = 0
        self.char_report_dict['accuracy_per_char'] = 0

    #edit distance
    def levenshteinDistance(self, s1, s2):
        if len(s1) > len(s2):
            s1, s2 = s2, s1

        distances = range(len(s1) + 1)
        for i2, c2 in enumerate(s2):
            distances_ = [i2+1]
            for i1, c1 in enumerate(s1):
                if c1 == c2:
                    distances_.append(distances[i1])
                else:
                    distances_.append(1 + min((distances[i1], distances[i1 + 1], distances_[-1])))
            distances = distances_
        return distances[-1]

    def __call__(self):
        super(OcrReport, self).__call__()
        assert self.char_report_dict['total_per_char'] > 0, 'total characters must grater than 0'
        self.char_report_dict['accuracy_per_char'] = self.char_report_dict['correct_per_char'] / self.char_report_dict['total_per_char']
        report_keys = [k for k,v in self.char_report_dict.items()]
        report_values = [v for k,v in self.char_report_dict.items() ]
        LOG.logI(self.char_report_str.format(*report_keys, *report_values))

# used in classifier
class ClassifierReport(Report):
    def __init__(self, ds_name="Unknown", total_num=0, cls_num=0, threshold=0):
        self.cls_num = cls_num
        self.enable = True
        if cls_num > 1000:
            warnings.warn("cls too much, disable confusion matrix and report module", UserWarning) 
            self.enable = False
        self.threshold = threshold
        Report.__init__(self, ds_name, total_num)

    def add(self, gt, pred):
        self.confusion_matrix[gt, pred] += 1
        return self
        
    def __call__(self):
        print(f"- DATASET: {self.ds_name}")
        if self.enable:
            self.calcReportDict()
            print(f"- ACCURACY: {self.accuracy:<.3f}")

            print("- CONFUSION-MATRIX")
            print(self.fmt_head)
            print(self.fmt_div)
            for i in range(self.cls_num):
                body = f"| cls{i} " + ("| {:<.0f} "*self.cls_num).format(*(self.confusion_matrix[i].toarray().flatten()))
                print(body)

            print("- TEST NSFW REPORT")
            print(self.fmt_head)
            print(self.fmt_div)
            for k, v in self.report_dict.items():
                body = f"| {k} " + ("| {:<.3f} "*self.cls_num).format(*v)
                print(body)

        indices, correlation = self.calcCorrelation()
        label, pred = indices
        idx = np.argsort(correlation)[::-1]
        correlation.sort()
        correlation = correlation[::-1]
        label, pred = label[idx], pred[idx]

        res = {}
        for i in range(label.size):
            if label[i] not in res:
                res[label[i]] = [] 
            res[label[i]].append((label[i], pred[i], round(correlation[i], 3)))

        print("- CORRELATION REPORT    ")
        for k, v in res.items():
            print(f"({k}: {v}    ")

        print(f"- DULATION: {time.time()-self.start_time :<.3f}s")

    def initReportFormat(self):
        self.fmt_head = (f"| {self.ds_name} " + "| cls{} " * self.cls_num).format(*range(self.cls_num))
        self.fmt_div = "|---" * (self.cls_num+1)

    def initMetricsDict(self):
        pass

    def reset(self):
        self.start_time = time.time()
        self.initReportDict()
        # self.confusion_matrix = np.zeros((self.cls_num, self.cls_num))
        # to save memory, init matrix use lil_matrix
        self.confusion_matrix = ss.lil_matrix((self.cls_num, self.cls_num))

    def initNumeratorKeys(self):
        pass

    def initReportDict(self):
        self.report_dict = OrderedDict()
        self.report_dict['precision'] = 0
        self.report_dict['recall'] = 0
        self.report_dict['f1-score'] = 0

    def initFuncDict(self):
        pass

    def calcReportDict(self):
        '''
        label -> left; pred -> top

        TP | FN
        ---|---
        FP | TN

        precision = TP / (TP+FP)
        recall = TP / (TP+FN)
        f1-score = 2TP/(2TP+FP+FN)
        '''
        confusion_matrix = self.confusion_matrix.tocsc()
        self.report_dict['recall'] = [confusion_matrix[i, i] / confusion_matrix[i].sum() for i in range(self.cls_num)]
        self.report_dict['precision'] = [confusion_matrix[i, i] / confusion_matrix[:, i].sum() for i in range(self.cls_num)]
        self.report_dict['f1-score'] = [(2*confusion_matrix[i, i]) / (confusion_matrix[i].sum()+confusion_matrix[:, i].sum()) for i in range(self.cls_num)]
        self.accuracy = confusion_matrix.diagonal().sum() / confusion_matrix.sum()

    def calcCorrelation(self):
        cls_total_num = (1/self.confusion_matrix.sum(axis=1)).reshape((-1, 1))
        i = list(range(self.cls_num))
        self.confusion_matrix[i, i] = 0
        # self.confusion_matrix /= cls_total_num
        self.confusion_matrix = self.confusion_matrix.multiply(cls_total_num)
        # self.confusion_matrix[self.confusion_matrix < self.threshold] = 0
        # indices = np.nonzero(self.confusion_matrix)
        indices = np.nonzero(self.confusion_matrix > self.threshold)
        correlation = self.confusion_matrix.tolil()[indices]
        return indices, correlation.toarray().flatten()


if __name__ == "__main__":
    print("==========test FaceReport============")
    report = FaceReport('gemfield',5)
    report.add("","2").add("gemfield","gemfield").add(1,1).add(None,None).add("1","1")
    report()

    print("==========test OcrReport============")
    report = OcrReport('gemfield',4)
    report.add('朝辞白帝彩云间', '朝辞白彩云间')
    report.add('君不见黄河之水天上来', '君不见黄河之水天上来')
    report.add('非汝之为美，美人之贻', '非汝之为美，美人之遗')
    report.add('gemfield', 'gem fie,ld')
    report()

    print("==========test ClassifierReport============")
    report = ClassifierReport('gemfield',5, 50000)
    report.add(0, 0).add(0, 0).add(0, 0).add(0, 0).add(0, 0).add(0, 0).add(0, 0).add(0, 0).add(0, 0).add(0, 3).\
            add(1, 0).add(1, 1).add(1, 1).add(1, 1).add(1, 1).add(1, 1).add(1, 1).add(1, 1).add(1, 2).\
            add(2, 1).add(2, 2).add(2, 2).add(2, 2).add(2, 2).add(2, 2).add(2, 3).add(2, 4).\
            add(3, 0).add(3, 3).add(3, 3).add(3, 3).add(3, 3).add(3, 3).add(3, 3).\
            add(4, 4).add(4, 4).add(4, 4).add(4, 4).add(4, 4).add(4, 4)
    report()



import cv2
import numpy as np
import random
from PIL import Image, ImageEnhance
from scipy import ndimage
from ..utils import LOG
from .warp_mls_helper import WarpMLS

class AugBase(object):
    def __init__(self, deepvac_aug_config):
        self.config = deepvac_aug_config
        self.auditConfig()

    def auditConfig(self):
        raise Exception("Not implemented!")

    def addUserConfig(self, config_name, user_give=None, developer_give=None, is_user_mandatory=False):
        if user_give is None and is_user_mandatory:
            LOG.logE("You must set mandatory config.aug.{} in config.py. Developer advised value:{}".format(config_name, developer_give),exit=True)
        if user_give:
            return user_give
        if developer_give:
            return developer_give
        LOG.logE("value missing for configuration: {}".format(config_name), exit=True)

    def __call__(self,img):
        raise Exception("Not implemented!")

    @staticmethod
    def auditInput(img, has_label=False):
        if not has_label:
            assert isinstance(img, np.ndarray) and img.ndim == 3, "image must be cv2 image"
            return

        if not isinstance(img, (list, tuple)):
            assert False, "parameter img must be [image, labels]"
            return

        assert len(img) == 2, "parameter img must be [image, labels]"
        img, label = img
        assert isinstance(img, np.ndarray) and img.ndim == 3, "image must be cv2 image"
        return img, label

    @staticmethod
    def pillow2cv(pillow_img, is_rgb2bgr=True):
        cv_image = np.array(pillow_img)
        if is_rgb2bgr:
            cv_image = cv2.cvtColor(cv_image, cv2.COLOR_RGB2BGR)
        return cv_image

    @staticmethod
    def cv2pillow(cv_img, is_bgr2rgb=True):
        if is_bgr2rgb:
            cv_img = cv2.cvtColor(cv_img, cv2.COLOR_BGR2RGB)
        return Image.fromarray(cv_img)

# 图像添加随机斑点
class SpeckleAug(AugBase):
    def __init__(self, deepvac_config):
        super(SpeckleAug,self).__init__(deepvac_config)

    def auditConfig(self):
        self.config.speckle_severity = self.addUserConfig('speckle_severity', self.config.speckle_severity, np.random.uniform(0, 0.6*255))

    def __call__(self, img):
        input_type = img.dtype
        self.auditInput(img)
        blur = ndimage.gaussian_filter(np.random.randn(*img.shape) * self.config.speckle_severity, 1)
        img_speck = (img + blur)
        img_speck[img_speck > 255] = 255
        img_speck[img_speck <= 0] = 0
        img_speck = img_speck.astype(input_type)
        return img_speck

# 仿射变换
class AffineAug(AugBase):
    def __init__(self, deepvac_config):
        super(AffineAug,self).__init__(deepvac_config)

    def auditConfig(self):
        # 空白填充色
        self.config.affine_borderValue = self.addUserConfig('affine_borderValue', self.config.affine_borderValue, (255,255,255))
        # x方向和y方向的伸缩率
        self.config.affine_shear_x = self.addUserConfig('affine_shear_x', self.config.affine_shear_x, 30)
        self.config.affine_shear_y = self.addUserConfig('affine_shear_y', self.config.affine_shear_y, 1)

    def __call__(self, img):
        self.auditInput(img)
        rows,cols=img.shape[:2]
        shear_x = float(np.random.randint(-self.config.affine_shear_x, self.config.affine_shear_x + 1))/100
        shear_y = float(np.random.randint(-self.config.affine_shear_y, self.config.affine_shear_y + 1))/100
        M = np.float32([[1.0, shear_x ,0.0],[shear_y, 1.0,0.0]])
        img_affine = cv2.warpAffine(img,M,(cols,rows),borderValue=self.config.affine_borderValue)
        return img_affine

# 透视变换
class PerspectAug(AugBase):
    def __init__(self, deepvac_config):
        super(PerspectAug,self).__init__(deepvac_config)

    def auditConfig(self):
        # b空白填充色
        self.config.perspect_borderValue = self.addUserConfig('perspect_borderValue', self.config.perspect_borderValue, (255,255,255))
        # 高h方向伸缩范围
        self.config.perspect_sh_lower = self.addUserConfig('perspect_sh_lower', self.config.perspect_sh_lower, 6)
        self.config.perspect_sh_upper = self.addUserConfig('perspect_sh_upper', self.config.perspect_sh_upper, 11)
        # 宽w方向伸缩范围
        self.config.perspect_sw_lower = self.addUserConfig('perspect_sw_lower', self.config.perspect_sw_lower, 20)
        self.config.perspect_sw_upper = self.addUserConfig('perspect_sw_upper', self.config.perspect_sw_upper, 31)

    def __call__(self, img):
        self.auditInput(img)
        h,w=img.shape[:2]
        scale_h = np.random.randint(self.config.perspect_sh_lower ,self.config.perspect_sh_upper)
        scale_w = np.random.randint(self.config.perspect_sw_lower ,self.config.perspect_sw_upper)
        point1 = np.array([[0,0],[w,0],[0,h],[w,h]],dtype = "float32")
        point2_list = [
            np.array([[w/scale_w,0],[(scale_w-1)*w/scale_w,0],[0,h],[w,h]],dtype = "float32"),
            np.array([[0,h/scale_h],[w,0],[0,(scale_h-1)*h/scale_h],[w,h]],dtype = "float32"),
            np.array([[0,0],[w,0],[w/scale_w,h],[(scale_w-1)*w/scale_w,h]],dtype = "float32"),
            np.array([[0,0],[w,h/scale_h],[0,h],[w,(scale_h-1)*h/scale_h]],dtype = "float32")]

        pt_idx = np.random.randint(0,4)
        M = cv2.getPerspectiveTransform(point1,point2_list[pt_idx])
        img_perspect = cv2.warpPerspective(img,M,(w,h),borderValue=self.config.perspect_borderValue)
        return img_perspect

# 高斯模糊
class GaussianAug(AugBase):
    def __init__(self, deepvac_config):
        super(GaussianAug,self).__init__(deepvac_config)

    def auditConfig(self):
        self.config.gaussian_ks = self.addUserConfig('gaussian_ks', self.config.gaussian_ks, [9,11,13,15,17])

    def __call__(self, img):
        self.auditInput(img)
        ks = self.config.gaussian_ks[np.random.randint(0,len(self.config.gaussian_ks))]
        img_gaussian = cv2.GaussianBlur(img,(ks, ks), 0)
        return img_gaussian

# 添加横线增强
class HorlineAug(AugBase):
    def __init__(self, deepvac_config):
        super(HorlineAug,self).__init__(deepvac_config)

    def auditConfig(self):
        # 线条间隔
        self.config.horline_space = self.addUserConfig('horline_space', self.config.horline_space, 4)
        # 线条颜色
        self.config.horline_color = self.addUserConfig('horline_color', self.config.horline_color, 0)
        # 线宽
        self.config.horline_thickness = self.addUserConfig('horline_thickness', self.config.horline_thickness, 1)

    def __call__(self, img):
        self.auditInput(img)
        img_h, img_w = img.shape[:2]
        for i in range(0,img_h,self.config.horline_space):
            cv2.line(img, (0, i), (img_w, i), self.config.horline_color, self.config.horline_thickness)
        return img

# 添加竖线增强
class VerlineAug(AugBase):
    def __init__(self, deepvac_config):
        super(VerlineAug,self).__init__(deepvac_config)

    def auditConfig(self):
        # 线条间隔
        self.config.verline_space = self.addUserConfig('verline_space', self.config.verline_space, 4)
        # 线条颜色
        self.config.verline_color = self.addUserConfig('verline_color', self.config.verline_color, 0)
        # 线宽
        self.config.verline_thickness = self.addUserConfig('verline_thickness', self.config.verline_thickness, 1)

    def __call__(self, img):
        self.auditInput(img)
        img_h, img_w = img.shape[:2]
        for i in range(0,img_w,self.config.verline_space):
            cv2.line(img, (i, 0), (i, img_h), self.config.verline_color, self.config.verline_thickness)
        return img

# 左右运动模糊
class LRmotionAug(AugBase):
    def __init__(self, deepvac_config):
        super(LRmotionAug,self).__init__(deepvac_config)

    def auditConfig(self):
        self.config.lrmotion_ks = self.addUserConfig('lrmotion_ks', self.config.lrmotion_ks, [3,5,7,9])

    def __call__(self,img):
        self.auditInput(img)
        ks = self.config.lrmotion_ks[np.random.randint(0,len(self.config.lrmotion_ks))]
        kernel_motion_blur = np.zeros((ks, ks))
        kernel_motion_blur[int((ks - 1) / 2), :] = np.ones(ks)
        kernel_motion_blur = kernel_motion_blur / ks
        img_lrmotion = cv2.filter2D(img, -1, kernel_motion_blur)
        return img_lrmotion

# 上下运动模糊
class UDmotionAug(AugBase):
    def __init__(self, deepvac_config):
        super(UDmotionAug,self).__init__(deepvac_config)

    def auditConfig(self):
        self.config.udmotion_ks = self.addUserConfig('udmotion_ks', self.config.udmotion_ks, [3,5,7,9])

    def __call__(self, img):
        self.auditInput(img)
        ks = self.config.udmotion_ks[np.random.randint(0,len(self.config.udmotion_ks))]
        kernel_motion_blur = np.zeros((ks, ks))
        kernel_motion_blur[:, int((ks - 1) / 2)] = np.ones(ks)
        kernel_motion_blur = kernel_motion_blur / ks
        img_udmotion = cv2.filter2D(img, -1, kernel_motion_blur)
        return img_udmotion

# 添加噪声
class NoisyAug(AugBase):
    def __init__(self, deepvac_config):
        super(NoisyAug,self).__init__(deepvac_config)

    def auditConfig(self):
        self.config.noisy_mean = self.addUserConfig('noisy_mean', self.config.noisy_mean, 0)
        self.config.noisy_sigma = self.addUserConfig('noisy_sigma', self.config.noisy_sigma, 1)

    def __call__(self, img):
        self.auditInput(img)
        row, col = img.shape[:2]
        gauss = np.random.normal(self.config.noisy_mean, self.config.noisy_sigma, (row, col,3))
        gauss = gauss.reshape(row, col,3)
        noisy = img + gauss
        img_noisy= noisy.astype(np.uint8)
        return img_noisy

# 扭曲变换
class DistortAug(AugBase):
    def __init__(self, deepvac_config):
        super(DistortAug,self).__init__(deepvac_config)

    def auditConfig(self):
        self.config.distort_segment = self.addUserConfig('distort_segment', self.config.distort_segment, 4)

    def __call__(self, img):
        self.auditInput(img)
        img_h, img_w = img.shape[:2]
        cut = img_w // self.config.distort_segment
        thresh = cut // 3
        if thresh == 0:
            return img

        src_pts = list()
        dst_pts = list()

        src_pts.append([0, 0])
        src_pts.append([img_w, 0])
        src_pts.append([img_w, img_h])
        src_pts.append([0, img_h])

        dst_pts.append([np.random.randint(thresh), np.random.randint(thresh)])
        dst_pts.append([img_w - np.random.randint(thresh), np.random.randint(thresh)])
        dst_pts.append([img_w - np.random.randint(thresh), img_h - np.random.randint(thresh)])
        dst_pts.append([np.random.randint(thresh), img_h - np.random.randint(thresh)])

        half_thresh = thresh * 0.5
        for cut_idx in np.arange(1, self.config.distort_segment, 1):
            src_pts.append([cut * cut_idx, 0])
            src_pts.append([cut * cut_idx, img_h])
            dst_pts.append([cut * cut_idx + np.random.randint(thresh) - half_thresh,
                            np.random.randint(thresh) - half_thresh])
            dst_pts.append([cut * cut_idx + np.random.randint(thresh) - half_thresh,
                            img_h + np.random.randint(thresh) - half_thresh])

        trans = WarpMLS(img, src_pts, dst_pts, img_w, img_h)
        img_distort = trans.generate()
        return img_distort

# 随机伸缩变换
class StretchAug(AugBase):
    def __init__(self, deepvac_config):
        super(StretchAug,self).__init__(deepvac_config)

    def auditConfig(self):
        self.config.stretch_segment = self.addUserConfig('stretch_segment', self.config.stretch_segment, 4)

    def __call__(self, img):
        self.auditInput(img)
        img_h, img_w = img.shape[:2]

        cut = img_w // self.config.stretch_segment
        thresh = cut * 4 // 5
        if thresh==0:
            return img

        src_pts = list()
        dst_pts = list()

        src_pts.append([0, 0])
        src_pts.append([img_w, 0])
        src_pts.append([img_w, img_h])
        src_pts.append([0, img_h])

        dst_pts.append([0, 0])
        dst_pts.append([img_w, 0])
        dst_pts.append([img_w, img_h])
        dst_pts.append([0, img_h])

        half_thresh = thresh * 0.5

        for cut_idx in np.arange(1, self.config.stretch_segment, 1):
            move = np.random.randint(thresh) - half_thresh
            src_pts.append([cut * cut_idx, 0])
            src_pts.append([cut * cut_idx, img_h])
            dst_pts.append([cut * cut_idx + move, 0])
            dst_pts.append([cut * cut_idx + move, img_h])

        trans = WarpMLS(img, src_pts, dst_pts, img_w, img_h)
        img_stretch = trans.generate()
        return img_stretch

# 透视变换
class PerspectiveAug(AugBase):
    def __init__(self, deepvac_config):
        super(PerspectiveAug,self).__init__(deepvac_config)

    def auditConfig(self):
        pass

    def __call__(self, img):
        self.auditInput(img)
        img_h, img_w = img.shape[:2]

        thresh = img_h // 2
        if thresh==0:
            return img

        src_pts = list()
        dst_pts = list()

        src_pts.append([0, 0])
        src_pts.append([img_w, 0])
        src_pts.append([img_w, img_h])
        src_pts.append([0, img_h])

        dst_pts.append([0, np.random.randint(thresh)])
        dst_pts.append([img_w, np.random.randint(thresh)])
        dst_pts.append([img_w, img_h - np.random.randint(thresh)])
        dst_pts.append([0, img_h - np.random.randint(thresh)])

        trans = WarpMLS(img, src_pts, dst_pts, img_w, img_h)
        img_perspective = trans.generate()
        return img_perspective

# 运动模糊
class MotionAug(AugBase):
    def __init__(self, deepvac_config):
        super(MotionAug, self).__init__(deepvac_config)

    def auditConfig(self):
        self.config.motion_degree = self.addUserConfig('motion_degree', self.config.motion_degree, 18)
        self.config.motion_angle = self.addUserConfig('motion_angle', self.config.motion_angle, 45)

    def __call__(self, img):
        self.auditInput(img)
        m = cv2.getRotationMatrix2D((self.config.motion_degree / 2, self.config.motion_degree / 2), self.config.motion_angle, 1)
        motion_blur_kernel = np.diag(np.ones(self.config.motion_degree))
        motion_blur_kernel = cv2.warpAffine(motion_blur_kernel, m, (self.config.motion_degree, self.config.motion_degree))
        motion_blur_kernel = motion_blur_kernel / self.config.motion_degree
        blurred = cv2.filter2D(img, -1, motion_blur_kernel)

        cv2.normalize(blurred, blurred, 0, 255, cv2.NORM_MINMAX)
        blurred = np.array(blurred, dtype=np.uint8)

        return blurred

# 降低图片亮度
class DarkAug(AugBase):
    def __init__(self, deepvac_config):
        super(DarkAug, self).__init__(deepvac_config)

    def auditConfig(self):
        self.config.dark_gamma = self.addUserConfig('dark_gamma', self.config.dark_gamma, 3)

    def __call__(self, img):
        self.auditInput(img)
        is_gray = img.ndim == 2 or img.shape[1] == 1
        if is_gray:
            img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
        hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
        illum = hsv[..., 2] / 255.
        illum = np.power(illum, self.config.dark_gamma)
        v = illum * 255.
        v[v > 255] = 255
        v[v < 0] = 0
        hsv[..., 2] = v.astype(np.uint8)
        img = cv2.cvtColor(hsv, cv2.COLOR_HSV2BGR)
        if is_gray:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

        return img

# 降低图片半边亮度
class HalfDarkAug(AugBase):
    def __init__(self, deepvac_config):
        super(HalfDarkAug, self).__init__(deepvac_config)

    def auditConfig(self):
        self.config.halfdark_gamma = self.addUserConfig('halfdark_gamma', self.config.halfdark_gamma, 1.5)

    def __call__(self, img):
        self.auditInput(img)
        h, w, _ = img.shape
        is_gray = img.ndim == 2 or img.shape[1] == 1
        if is_gray:
            img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
        hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
        illum = hsv[..., 2] / 255.
        illum[:, w//2:] = np.power(illum[:, w//2:], self.config.halfdark_gamma)
        v = illum * 255
        v[v > 255] = 255
        v[v < 0] = 0
        hsv[..., 2] = v.astype(np.uint8)
        img = cv2.cvtColor(hsv, cv2.COLOR_HSV2BGR)
        if is_gray:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

        return img

# 模拟IPC场景增强
class IPCFaceAug(AugBase):
    def __init__(self, deepvac_config):
        super(IPCFaceAug, self).__init__(deepvac_config)
        self.deepvac_config = deepvac_config

    def auditConfig(self):
        pass

    def __call__(self, img):
        self.auditInput(img)
        half_dark = HalfDarkAug(self.deepvac_config)
        half_darked = half_dark(img)
        motion = MotionAug(self.deepvac_config)
        motioned = motion(half_darked)
        return motioned

# 随机crop框降低亮度
class RandomCropDarkAug(AugBase):
    def __init__(self, deepvac_config):
        super(RandomCropDarkAug, self).__init__(deepvac_config)

    def auditConfig(self):
        self.config.random_crop_dark_gamma = self.addUserConfig('random_crop_dark_gamma', self.config.random_crop_dark_gamma, 1.2)

    def __call__(self, img):
        self.auditInput(img)
        height, width, _ = img.shape
        w = np.random.uniform(0.3 * width, width)
        h = np.random.uniform(0.3 * height, height)
        if h / w < 0.5 or h / w > 2:
            return img
        left = np.random.uniform(width - w)
        top = np.random.uniform(height - h)

        rect = np.array([int(left), int(top), int(left+w), int(top+h)])
        current_img = img[rect[1]:rect[3], rect[0]:rect[2], :]
        hsv = cv2.cvtColor(current_img, cv2.COLOR_BGR2HSV)
        illum = hsv[..., 2] / 255.
        illum = np.power(illum, self.config.random_crop_dark_gamma)
        v = illum * 255.
        v[v > 255] = 255
        v[v < 0] = 0
        hsv[..., 2] = v.astype(np.uint8)
        dark_img = cv2.cvtColor(hsv, cv2.COLOR_HSV2BGR)

        for x in range(rect[1], rect[3]):
            for y in range (rect[0], rect[2]):
                img[x, y, 0] = dark_img[x-rect[1], y-rect[0], 0]
                img[x, y, 1] = dark_img[x-rect[1], y-rect[0], 1]
                img[x, y, 2] = dark_img[x-rect[1], y-rect[0], 2]

        LOG.logI('rect in RandomCropDarkAug:{}'.format(rect))
        return img

# 随机颜色扰动
class ColorJitterAug(AugBase):
    def __init__(self, deepvac_config):
        super(ColorJitterAug, self).__init__(deepvac_config)

    def auditConfig(self):
        pass

    def __call__(self, img):
        self.auditInput(img)
        img = self.cv2pillow(img)
        img = ImageEnhance.Color(img).enhance(np.random.uniform(0.8, 1.3))
        return self.pillow2cv(img)

class BrightnessJitterAug(AugBase):
    def __init__(self, deepvac_config):
        super(BrightnessJitterAug, self).__init__(deepvac_config)

    def auditConfig(self):
        pass

    def __call__(self, img):
        self.auditInput(img)
        img = self.cv2pillow(img)
        img = ImageEnhance.Brightness(img).enhance(np.random.uniform(0.6, 1.5))
        return self.pillow2cv(img)

class ContrastJitterAug(AugBase):
    def __init__(self, deepvac_config):
        super(ContrastJitterAug, self).__init__(deepvac_config)

    def auditConfig(self):
        pass

    def __call__(self, img):
        self.auditInput(img)
        img = self.cv2pillow(img)
        img = ImageEnhance.Contrast(img).enhance(np.random.uniform(0.5, 1.8))
        return self.pillow2cv(img)

class RandomColorJitterAug(AugBase):
    def __init__(self, deepvac_config):
        super(RandomColorJitterAug, self).__init__(deepvac_config)

    def auditConfig(self):
        pass

    def __call__(self, img):
        self.auditInput(img)
        img = self.cv2pillow(img)
        if random.randint(0, 1):
            img = ImageEnhance.Color(img).enhance(np.random.uniform(0.8, 1.3))
        if random.randint(0, 1):
            img = ImageEnhance.Brightness(img).enhance(np.random.uniform(0.6, 1.5))
        if random.randint(0, 1):
            img = ImageEnhance.Contrast(img).enhance(np.random.uniform(0.5, 1.8))
        return self.pillow2cv(img)

class MosaicAug(AugBase):
    def __init__(self, deepvac_config):
        super(MosaicAug, self).__init__(deepvac_config)
        self.neighbor = 2

    def auditConfig(self):
        pass

    def __call__(self, img):
        self.auditInput(img)
        neighbor = self.neighbor
        h, w = img.shape[0], img.shape[1]
        for i in range(0, h - neighbor -1, neighbor):
            for j in range(0, w - neighbor - 1, neighbor):
                rect = [j, i, neighbor, neighbor]
                color = img[i][j].tolist()
                left_up = (rect[0], rect[1])
                right_down = (rect[0] + neighbor - 1, rect[1] + neighbor - 1)
                cv2.rectangle(img, left_up, right_down, color, -1)
        return img



import sys
from torchvision import transforms as trans
from ..utils import LOG
from ..core.factory import SyszuxFactory
from . import base_aug, face_aug, seg_aug, text_aug, yolo_aug

locals = locals()
class AugFactory(SyszuxFactory):
    def __init__(self):
        super(AugFactory,self).__init__()

    def auditConfig(self):
        super(AugFactory,self).auditConfig()
        self.end_suffix = 'Aug'
        self.locals = locals

    def addProduct(self, name, ins):
        self.factory_dict[name] = eval(ins)

    def initProducts(self):
        self.addProducts(base_aug, product_name='base_aug', end_suffix='Aug')
        self.addProducts(face_aug, product_name='face_aug', end_suffix='Aug')
        self.addProducts(seg_aug, product_name='seg_aug', end_suffix='Aug')
        self.addProducts(text_aug, product_name='text_aug', end_suffix='Aug')
        self.addProducts(yolo_aug, product_name='yolo_aug', end_suffix='Aug')
        self.addProducts(trans, product_name='trans', omit_list=["Compose","transforms","functional"])

    def initMyProducts(self):
        pass
        
if __name__ == "__main__":
    aug = AugFactory()
    LOG.logI(aug.factory_dict)



import torch.nn as nn
from .module_helper import makeDivisible
from .conv_layer import Conv2dBNReLU, Conv2dBNHswish
from .se_layer import SELayer
from .activate_layer import hswish

class ResBnBlock(nn.Module):
     def __init__(self, in_channel, expand_channel, out_channel, kernel_size, stride, shortcut=True):
         super(ResBnBlock, self).__init__()
         self.layer = nn.Sequential(
                 Conv2dBNReLU(in_channel, expand_channel, 1, 1),
                 Conv2dBNReLU(expand_channel, expand_channel, kernel_size, stride, kernel_size//2, expand_channel),
                 Conv2dBN(expand_channel, out_channel, 1, 1),
                 )
         self.shortcut = shortcut and (in_channel == out_channel)

     def forward(self, x):
         return self.layer(x) + x if self.shortcut else self.layer(x)

class InvertedResidual(nn.Module):
    def __init__(self, inp, oup, kernel_size, stride, expand_ratio, use_se, use_hs, padding=None):
        super(InvertedResidual, self).__init__()
        if padding is None:
            padding = (kernel_size - 1) // 2
        hidden_dim = makeDivisible(inp * expand_ratio, 8)
        assert stride in [1, 2, (2, 1)]
        assert kernel_size in [3,5]

        self.use_res_connect = stride == 1 and inp == oup
        layers = []
        if expand_ratio != 1:
            layers.append(Conv2dBNHswish(inp, hidden_dim, kernel_size=1))  # pw

        layers.extend([
            # dw
            nn.Conv2d(hidden_dim, hidden_dim, kernel_size, stride, padding, groups=hidden_dim, bias=False),
            nn.BatchNorm2d(hidden_dim),
            hswish() if use_hs else nn.ReLU(inplace=True),
            # Squeeze-and-Excite
            SELayer(hidden_dim) if use_se else nn.Identity(),
            # pw-linear
            nn.Conv2d(hidden_dim, oup, kernel_size=1),
            nn.BatchNorm2d(oup),
        ])
        self.conv = nn.Sequential(*layers)

    def forward(self, x):
        if self.use_res_connect:
            return x + self.conv(x)
        else:
            return self.conv(x)

class InvertedResidualFacialKpBlock(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, expansion_factor: int=6, kernel_size: int=3, stride: int=2, padding: int=1,
                 is_residual: bool=True):
        super(InvertedResidualFacialKpBlock, self).__init__()
        assert stride in [1, 2]

        self.block = nn.Sequential(
            nn.Conv2d(in_channels, in_channels * expansion_factor, 1, bias=False),
            nn.BatchNorm2d(in_channels * expansion_factor),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels * expansion_factor, in_channels * expansion_factor,
                      kernel_size, stride, padding, 1,
                      groups=in_channels * expansion_factor, bias=False),
            nn.BatchNorm2d(in_channels * expansion_factor),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels * expansion_factor, out_channels, 1,
                      bias=False),
            nn.BatchNorm2d(out_channels))

        self.is_residual = is_residual if stride == 1 else False
        self.is_conv_res = False if in_channels == out_channels else True
        #todo
        if stride == 1 and self.is_conv_res:
            self.conv_res = nn.Sequential(nn.Conv2d(in_channels, out_channels, 1, bias=False),
                                          nn.BatchNorm2d(out_channels))

    def forward(self, x):
        block = self.block(x)
        if self.is_residual:
            if self.is_conv_res:
                return self.conv_res(x) + block
            return x + block
        return block


from torchvision.datasets import ImageFolder

#Dataset -> VisionDataset -> DatasetFolder -> ImageFolder -> *Dataset
class ImageFolderWithTransformDataset(ImageFolder):
    def __init__(self, deepvac_config):
        self.transform_op = deepvac_config.transform_op
        self.img_folder = deepvac_config.img_folder
        super(ImageFolderWithTransformDataset,self).__init__(self.img_folder, self.transform_op)

class ImageFolderWithPathsDataset(ImageFolderWithTransformDataset):
    def __init__(self, deepvac_config):
        super(ImageFolderWithPathsDataset, self).__init__(deepvac_config)
    # override the __getitem__ method. this is the method that dataloader calls
    def __getitem__(self, index):
        # this is what ImageFolder normally returns
        original_tuple = super(ImageFolderWithPathsDataset, self).__getitem__(index)
        # the image file path
        path = self.imgs[index][0]
        # make a new tuple that includes original and the path
        tuple_with_path = (original_tuple + (path,))
        #print("gemfieldimg: ", tuple_with_path[0],'\t',tuple_with_path[1],'\t',tuple_with_path[2])
        return tuple_with_path


import os
import subprocess
from ..utils import LOG
from .base import DeepvacCast

class NcnnCast(DeepvacCast):
    def __init__(self, deepvac_train_config):
        super(NcnnCast,self).__init__(deepvac_train_config)

    def auditConfig(self):
        if not self.config.ncnn_model_dir:
            return False

        if not self.config.onnx2ncnn:
            LOG.logE("You must set the onnx2ncnn executable program path in config file. If you want to compile onnx2ncnn tools, reference https://github.com/Tencent/ncnn/wiki/how-to-build#build-for-linux-x86 ", exit=True)
            return False
        return True

    def process(self, cast_output_file=None):
        output_ncnn_file = self.config.ncnn_model_dir
        if cast_output_file:
            output_ncnn_file = '{}/ncnn__{}.bin'.format(self.config.output_dir, cast_output_file)
            self.config.ncnn_model_dir = output_ncnn_file
        self.config.ncnn_arch_dir = '{}.param'.format(output_ncnn_file)
        LOG.logI("config.ncnn_model_dir found, save ncnn model to {}...".format(self.config.ncnn_model_dir))

        #to onnx, also set self.config.onnx_model_dir, self.config.onnx_input_names and self.config.onnx_output_names
        self.exportOnnx()

        cmd = self.config.onnx2ncnn + " " + self.config.onnx_model_dir + " " + self.config.ncnn_arch_dir + " " + output_ncnn_file
        pd = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if pd.stderr.read() == b"":
            LOG.logI("Pytorch model convert to NCNN model succeed, save ncnn param file in {}, save ncnn bin file in {}.".format(self.config.ncnn_arch_dir, output_ncnn_file))
            return

        LOG.logE(pd.stderr.read() + b". Error occured when export ncnn model. Deepvac try to simplify the model first.")
        try:
            import onnx
            from onnxsim import simplify
        except:
            LOG.logE("You must install onnx and onnxsim package first if you want to convert pytorch to ncnn.", exit=True)

        model_op, check_ok = simplify(self.config.onnx_model_dir, check_n=3, perform_optimization=True, skip_fuse_bn=True,  skip_shape_inference=False)
        onnx.save(model_op, self.config.onnx_model_dir)
        if not check_ok:
            LOG.logE("Maybe something wrong when simplify the model, we can't guarantee generate model is right.")
        else:
            LOG.logI("Simplify model succeed")
        subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if pd.stderr.read() != b"":
            LOG.logE(pd.stderr.read() + b". we can't guarantee generate model is right.", exit=True)
        
        if not os.path.isfile(output_ncnn_file):
            LOG.logE("Error: ncnn model not generated due to internal error!", exit=True)

        LOG.logI("Pytorch model convert to NCNN model succeed, save ncnn param file in {}, save ncnn bin file in {}".format(self.config.ncnn_arch_dir, output_ncnn_file))



class AttrDict(dict):
    def __getattr__(self, key):
        return self.get(key)

    def __setattr__(self, key, value):
        if key in self.__dict__:
            self.__dict__[key] = value
        else:
            self[key] = value

config = AttrDict()
config.train = AttrDict()
config.feature = AttrDict()
config.aug = AttrDict()

## ------------------ common ------------------
config.train.device = "cuda"
config.train.output_dir = "output"
config.train.log_dir = "log"
config.train.log_every = 10
config.train.disable_git = False
config.train.cast2cpu = True

## ------------------ ddp --------------------
config.train.dist_url = "tcp://localhost:27030"
config.train.world_size = 2

## ------------------ optimizer  ------------------
config.train.lr = 0.01
config.train.lr_step = None
config.train.lr_factor = 0.2703
config.train.momentum = 0.9
config.train.nesterov = False
config.train.weight_decay = None

## -------------------- loader ------------------
config.train.num_workers = 3
config.train.nominal_batch_factor = 1

## ------------------- train ------------------
config.train.train_batch_size = 128
config.train.epoch_num = 30
#model save number duriong an epoch
config.train.save_num = 5
config.train.milestones = [2,4,6,8,10]
config.train.checkpoint_suffix = ''

## ------------------ val ------------------
config.train.val_batch_size = None




import os
import sys
import numpy as np
import torch
import time
try:
    import faiss
except:
    LOG.W("Faiss not installed. You cannot use the API implemented based on Faiss library.")

from deepvac.syszux_report import ClassifierReport
from deepvac.syszux_log import LOG

def swigPtrFromTensor(x):
    """ gets a Faiss SWIG pointer from a pytorch trensor (on CPU or GPU) """
    assert x.is_contiguous()

    if x.dtype == torch.float32:
        return faiss.cast_integer_to_float_ptr(x.storage().data_ptr() + x.storage_offset() * 4)
    
    if x.dtype == torch.int64:
        return faiss.cast_integer_to_long_ptr(x.storage().data_ptr() + x.storage_offset() * 8)

    raise Exception("tensor type not supported: {}".format(x.dtype))

#from config import config
#config.feature as feature_vector_config
class FeatureVector(object):
    def __init__(self, feature_vector_config):
        self.config = feature_vector_config
        assert self.config.db_path_list, 'You must configure config.db_path_list in config.py'
        assert self.config.map_locs, 'You must configure config.map_locs in config.py, which is to specify how mapping the feature to CUDA/CPU device'
        assert len(self.config.map_locs) == len(self.config.db_path_list), 'config.db_path_list number should be same with config.map_locs'
        self.dbs = []
        if not self.config.name:
            self.config.name = 'gemfield'

        if not self.config.log_every:
            self.config.log_every = 10000

    def loadDB(self):
        for db_path, map_loc in zip(self.config.db_path_list, self.config.map_locs):
            self.dbs.append(torch.load(db_path, map_location = map_loc))

    def searchDB(self, xb, xq, k=2):
        D = []
        I = []
        if k < 1 or k > 10:
            LOG.logE('illegal nearest neighbors parameter k(1 ~ 10): {}'.format(k))
            return D, I

        distance = torch.norm(xb - xq, dim=1)
        for i in range(k):
            values, indices = distance.kthvalue(i+1)
            D.append(values.item())
            I.append(indices.item())
        return D, I

    def searchAllDB(self, xq, k=2):
        raise Exception("Cannot do searchAllDB in base class, since no names can be used to re index feature accross multi DB BLOCK.")

    def printClassifierReport(self):
        pass

class NamesPathsClsFeatureVector(FeatureVector):
    def __init__(self, feature_vector_config):
        super(NamesPathsClsFeatureVector, self).__init__(feature_vector_config)
        assert self.config.class_num, 'You must configure config.class_num in config.py'
        assert self.config.np_path_list, 'You must configure config.np_path_list in config.py'
        assert len(self.config.np_path_list) == len(self.config.db_path_list), 'config.db_path_list number should be same with config.np_path_list'
        self.names = []
        self.paths = []

    def loadDB(self):
        super(NamesPathsClsFeatureVector, self).loadDB()
        for np_path in self.config.np_path_list:
            npf = np.load(np_path)
            self.names.append(npf['names'])
            self.paths.append(npf['paths'])

    def searchAllDB(self, xq, k=2):
        TMP_D = []
        TMP_N = []
        for i, db in enumerate(self.dbs):
            xq = xq.to(db.device)
            distances, indices = self.searchDB(db, xq, k)
            TMP_D.extend(distances)
            TMP_N.extend(self.names[i][indices])

        N = [n for _,n in sorted(zip(TMP_D,TMP_N))]
        D = sorted(TMP_D)

        return D[:k], N[:k]

    def printClassifierReport(self):
        report = ClassifierReport(self.config.name, self.config.class_num, self.config.class_num)
        for i, db in enumerate(self.dbs):
            name = self.names[i]
            path = self.paths[i]
            for idx, emb in enumerate(db):
                _, N = self.searchAllDB(emb)
                should_be_gt, pred = N[0], N[1]
                LOG.logI("label : {}".format(name[idx]))
                LOG.logI("pred : {}".format(pred))
                if idx % self.config.log_every == 0:
                    LOG.logI("Process {} img...".format(idx))
                report.add(int(name[idx]), int(pred))
        report()

class NamesPathsClsFeatureVectorByFaiss(object):
    def __init__(self, feature_vector_config):
        if 'faiss' not in sys.modules:
            LOG.logE("Faiss not installed, cannot use this class.", exit=True)
        self.config = feature_vector_config
        assert self.config.db_path_list, 'You must configure config.db_path_list in config.py'
        assert self.config.np_path_list, 'You must configure config.np_path_list in config.py'
        assert len(self.config.np_path_list) == len(self.config.db_path_list), 'config.np_path_list number should be same with config.db_path_list'
                        
        if not self.config.name:
            self.config.name = 'gemfield'
        if not self.config.log_every:
            self.config.log_every = 10000
        
        self.dbs = []
        self.names = []
        self.paths = []
        self.gpu_index = []

    def loadDB(self):
        for db_path, np_path in zip(self.config.db_path_list, self.config.np_path_list):
            npf = np.load(np_path)
            self.dbs = np.vstack((self.dbs, torch.load(db_path).to('cpu').numpy().astype('float32'))) \
                    if len(self.dbs) else torch.load(db_path).to('cpu').numpy().astype('float32')
            self.names = np.hstack((self.names, npf['names'])) if len(self.names) else npf['names']
            self.paths = np.hstack((self.paths, npf['paths'])) if len(self.paths) else npf['paths']
        torch.cuda.empty_cache()
        self.__loadIndex()

    def __loadIndex(self):
        assert self.dbs != [], "You should load db before load index, use self.loadDB() ..."
        d = self.dbs[0].shape[-1]
        ngpu = faiss.get_num_gpus()
        index = faiss.IndexFlatL2(d)
        vres = faiss.GpuResourcesVector()
        vdev = faiss.IntVector()
        gpu_resources = []
        
        for i in range(0, ngpu):
            res = faiss.StandardGpuResources()
            gpu_resources.append(res)
            vdev.push_back(i)
            vres.push_back(res)
        
        co = faiss.GpuMultipleClonerOptions()
        co.shard = True
        self.gpu_index = faiss.index_cpu_to_gpu_multiple(vres, vdev, index, co)
        self.gpu_index.referenced_objects = gpu_resources
        self.gpu_index.add(self.dbs)

    def searchAllDB(self, xq, k=2):
        assert self.gpu_index != [], "You should load index before search, use self.loadIndex() ..."
        D, I = self.gpu_index.search(xq, k)
        N = self.names[I]

        return D, N

    def printClassifierReport(self):
        report = ClassifierReport(self.config.name, self.config.class_num, self.config.class_num)
        RD, RN = self.searchAllDB(self.dbs)

        for idx, n in enumerate(RN):
            pred_ori, pred = n[0], n[1]
            LOG.logI("label : {}".format(self.names[idx]))
            LOG.logI("pred : {}".format(pred))
            if idx % self.config.log_every == 0:
                LOG.logI("Process {} img...".format(idx))
            
            report.add(int(self.names[idx]), int(pred))
        report()

class NamesPathsClsFeatureVectorByFaissMulBlock(NamesPathsClsFeatureVectorByFaiss):
    def __init__(self, feature_vector_config):
        super(NamesPathsClsFeatureVectorByFaissMulBlock, self).__init__(feature_vector_config)
        assert len(self.config.np_path_list) == len(self.config.block_map), "config.np_path_list number should be same with config.block_map"
        self.dbs = [[]] * faiss.get_num_gpus()
        self.names = [[]] * faiss.get_num_gpus()
        self.paths = [[]] * faiss.get_num_gpus()

    def loadDB(self):
        for block_map, db_path, np_path in zip(self.config.block_map, self.config.db_path_list, self.config.np_path_list):
            npf = np.load(np_path)
            self.dbs[block_map] = np.vstack((self.dbs[block_map], torch.load(db_path).to('cpu').numpy().astype('float32'))) \
                                    if len(self.dbs[block_map]) else torch.load(db_path).to('cpu').numpy().astype('float32')
            self.names[block_map] = np.hstack((self.names[block_map], npf['names'])) if len(self.names[block_map]) else npf['names']
            self.paths[block_map] = np.hstack((self.paths[block_map], npf['paths'])) if len(self.paths[block_map]) else npf['paths']
        
        
        torch.cuda.empty_cache()
        self.__loadIndex()

    def __loadIndex(self):
        assert self.dbs != [], "You should load db before load index, use self.loadDB() ..."
        d = self.dbs[0].shape[-1]
        ngpu = faiss.get_num_gpus()
        index = faiss.IndexFlatL2(d)

        res = faiss.StandardGpuResources()
        
        for i, db in enumerate(self.dbs):
            gpu_index = faiss.index_cpu_to_gpu(res, i, index)
            gpu_index.add(db)
            self.gpu_index.append(gpu_index)

    def searchAllDB(self, xq, k=2):
        assert self.gpu_index != [], "You should load index before search, use self.loadIndex() ..."
        
        D = []
        N = []
        for i, gindex in enumerate(self.gpu_index):
            tempD, tempI = gindex.search(xq, k)
            tempN = self.names[i][tempI]
            if i == 0:
                D, N = tempD, tempN
                continue
            D = np.hstack((D, tempD))
            N = np.hstack((N, tempN))

        RD, RN = D, N
        for i, d in enumerate(D):
            index = np.argsort(d)
            RD[i] = D[i][index]
            RN[i] = N[i][index]
            
        return RD[:, :k], RN[:, :k]

    def printClassifierReport(self):
        report = ClassifierReport(self.config.name, self.config.class_num, self.config.class_num)
        
        for i, db in enumerate(self.dbs):
            name = self.names[i]
            RD, RN = self.searchAllDB(db)

            for idx, n in enumerate(RN):
                pred_ori, pred = n[0], n[1]
                LOG.logI("label : {}".format(name[idx]))
                LOG.logI("pred : {}".format(pred))
                if idx % self.config.log_every == 0:
                    LOG.logI("Process {} img...".format(idx))
                
                report.add(int(name[idx]), int(pred))
            LOG.logI("{} xq finished...".format(i))
                
        report()

class NamesPathsClsFeatureVectorByFaissPytorch(NamesPathsClsFeatureVector):
    def __init__(self, feature_vector_config):
        super(NamesPathsClsFeatureVectorByFaissPytorch, self).__init__(feature_vector_config)
        self.dbs = []
                           
    def loadDB(self):
        super(NamesPathsClsFeatureVectorByFaissPytorch, self).loadDB()
        assert self.config.db_path_list, 'You must configure config.db_path_list in config.py'
        assert self.config.map_locs, 'You must configure config.map_locs in config.py'
        for db_path, map_loc in zip(self.config.db_path_list, self.config.map_locs):
            self.dbs.append(torch.load(db_path, map_location = map_loc))

    def search_index_pytorch(self, index, x, k, D=None, I=None):
        """call the search function of an index with pytorch tensor I/O (CPU
            and GPU supported)"""
        assert x.is_contiguous()
        n, d = x.size()
        assert d == index.d
        
        if D is None:
            D = torch.empty((n, k), dtype=torch.float32, device=x.device)             
        else:                                        
            assert D.size() == (n, k)
        
        if I is None:
            I = torch.empty((n, k), dtype=torch.int64, device=x.device)             
        else:                                                           
            assert I.size() == (n, k)
        
        torch.cuda.synchronize()
        xptr = swigPtrFromTensor(x)
        Iptr = swigPtrFromTensor(I)
        Dptr = swigPtrFromTensor(D)
        index.search_c(n, xptr, k, Dptr, Iptr)
        torch.cuda.synchronize()
        
        return D, I

    def searchAllDB(self, xq, k=2):        
        D = []
        N = []
        d = 512
        LOG.logI('Load index start...')
        res = faiss.StandardGpuResources()
        gpu_index = faiss.GpuIndexFlatL2(res, d)
        LOG.logI('Load index finished...')

        for i, db in enumerate(self.dbs):
            name = self.names[i]
            gpu_index.reset()
            gpu_index.add(db.to('cpu').numpy().astype('float32'))
            tempD, tempI = self.search_index_pytorch(gpu_index, xq, k)
            tempD = tempD.to('cpu').numpy()
            tempI = tempI.to('cpu').numpy()
            tempN = name[tempI]
            
            if len(D) == 0:
                D, N = tempD, tempN
                continue

            D = np.hstack((D, tempD))
            N = np.hstack((N, tempN))
            
        RD, RN = D, N    
        for i, d in enumerate(D):
            index = np.argsort(d)
            RD[i] = D[i][index]
            RN[i] = N[i][index]
        
        return RD[:, :k], RN[:, :k]

    def printClassifierReport(self):        
        report = ClassifierReport(self.config.name, self.config.class_num, self.config.class_num)
        for i, xq in enumerate(self.dbs):
            LOG.logI('Process {} db...'.format(i))
            name = self.names[i]
            path = self.paths[i]
            RD, RN = self.searchAllDB(xq)
            
            for idx, n in enumerate(RN):
                pred_ori, pred = n[0], n[1]
                LOG.logI("label : {}".format(name[idx]))
                LOG.logI("pred : {}".format(pred))
                if idx % self.config.log_every == 0:
                    LOG.logI("Process {} img...".format(idx))
                
                report.add(int(name[idx]), int(pred))
        
        report()

if __name__ == "__main__":
    from config import config
    fv = NamesPathsClsFeatureVector(config)
    fv.printClassifierReport()

def loadDB(self, db_path):
    self.xb = torch.load(db_path).to(self.device)

def addEmb2DB(self, emb):
    self.xb = torch.cat((self.xb, emb))

def saveDB(self, db_path):
    torch.save(self.xb, db_path)

def search(self, xq, k=1):
    D = []
    I = []
    if k < 1 or k > 10:
        LOG.logE('illegal nearest neighbors parameter k(1 ~ 10): {}'.format(k))
        return D, I

    distance = torch.norm(self.xb - xq, dim=1)
    for i in range(k):
        values, indices = distance.kthvalue(i+1)
        D.append(values.item())
        I.append(indices.item())
    return D, I


import torch
import torch.nn as nn

class SPP(nn.Module):
    # Spatial pyramid pooling layer used in YOLOv3-SPP
    def __init__(self, in_planes, out_planes, pool_kernel_size=(5, 9, 13), eps=1e-5, momentum=0.1):
        super(SPP, self).__init__()
        hidden_planes = in_planes // 2  # hidden channels
        self.conv1 = Conv2dBNHardswish(in_planes, hidden_planes, 1, 1, eps, momentum)
        self.conv2 = Conv2dBNHardswish(hidden_planes * (len(pool_kernel_size) + 1), out_planes, 1, 1, eps, momentum)
        self.pool_list = nn.ModuleList([nn.MaxPool2d(kernel_size=k, stride=1, padding=k // 2) for k in pool_kernel_size])

    def forward(self, x):
        x = self.conv1(x)
        return self.conv2(torch.cat([x] + [pool(x) for pool in self.pool_list], 1))






import os
import numpy as np
import cv2
from torch.utils.data import Dataset

class CocoCVDataset(Dataset):
    def __init__(self, deepvac_config):
        super(CocoCVDataset, self).__init__()
        self.conf = deepvac_config
        self.transform = deepvac_config.transform
        self.img_folder = deepvac_config.img_folder
        self.aug_composer = deepvac_config.aug_composer

        try:
            from pycocotools.coco import COCO
        except:
            raise Exception("pycocotools module not found, you should try 'pip3 install pycocotools' first!")
        self.coco = COCO(deepvac_config.annotation)
        self.ids = list(sorted(self.coco.imgs.keys()))
        self.cats = list(sorted(self.coco.cats.keys()))

    def __len__(self):
        return len(self.ids)

    def __getitem__(self, index):
        sample, label = self._getSample(index)
        # augment for numpy
        if self.aug_composer:
            sample = self.aug_composer(sample)
        # pytorch offical transform for PIL Image or numpy
        if self.transform is not None:
            # rewrite method: '_buildSample' first
            sample = self.transform(sample)
        # post process
        img, label = self._buildSample(sample, label)
        return img, label

    def _getSample(self, index):
        img = self.loadImgs(index)
        category_ids, boxes, masks = self.loadAnns(index)
        # return target you want
        return img, category_ids

    def _buildSample(self, img, target):
        # BGR -> RGB && (H, W, C) -> (C, H, W)
        img = img[:, :, ::-1].transpose(2, 0, 1)
        img = np.ascontiguousarray(img)
        # numpy -> tensor
        img = torch.from_numpy(img)
        target = torch.from_numpy(target)
        return img, target

    def loadImgs(self, index):
        file_name = self.coco.loadImgs(self.ids[index])[0]['file_name']
        img = cv2.imread(os.path.join(self.img_folder, file_name), 1)
        assert img is not None, "Image {} not found!".format(file_name)
        return img

    def loadAnns(self, index):
        ann_ids = self.coco.getAnnIds(imgIds=self.ids[index])
        anns = self.coco.loadAnns(ann_ids)
        category_ids = np.array([self.cats.index(i["category_id"]) for i in anns], dtype=np.float)
        boxes = np.array([i["bbox"] for i in anns], dtype=np.float)
        masks = np.array([self.coco.annToMask(i) for i in anns], dtype=np.float)
        return category_ids, boxes, masks



import math
import cv2
import numpy as np
import random
from .base_aug import AugBase

### yolov5 dataset aug
class HSVAug(AugBase):
    def __init__(self, deepvac_config):
        super(HSVAug, self).__init__(deepvac_config)

    def auditConfig(self):
        self.config.hsv_hgain = self.addUserConfig('dark_gamma', self.config.dark_gamma, 0.015)
        self.config.hsv_sgain = self.addUserConfig('dark_gamma', self.config.dark_gamma, 0.7)
        self.config.hsv_vgain = self.addUserConfig('dark_gamma', self.config.dark_gamma, 0.4)

    def __call__(self, img):
        img, label = self.auditInput(img, has_label=True)
        assert isinstance(label, np.ndarray) and label.ndim == 2, "label must be numpy.ndarray, and shape should be (n, 5)"
        hue, sat, val = cv2.split(cv2.cvtColor(img, cv2.COLOR_BGR2HSV))
        dtype = img.dtype
        x = np.arange(0, 256, dtype=np.int16)
        # 随机增幅
        r = np.random.uniform(-1, 1, 3) * [self.config.hsv_hgain, self.config.hsv_sgain, self.config.hsv_vgain] + 1
        lut_hue = ((x * r[0]) % 180).astype(dtype)
        lut_sat = np.clip(x * r[1], 0, 255).astype(dtype)
        lut_val = np.clip(x * r[2], 0, 255).astype(dtype)
        img_hsv = cv2.merge((cv2.LUT(hue, lut_hue), cv2.LUT(sat, lut_sat), cv2.LUT(val, lut_val))).astype(dtype)
        cv2.cvtColor(img_hsv, cv2.COLOR_HSV2BGR, dst=img)
        return img, label

class YoloHFlipAug(AugBase):
    def __init__(self, deepvac_config):
        super(YoloHFlipAug, self).__init__(deepvac_config)

    def auditConfig(self):
        pass

    def __call__(self, img):
        img, label = self.auditInput(img, has_label=True)
        assert isinstance(label, np.ndarray) and label.ndim == 2, "label must be numpy.ndarray, and shape should be (n, 5)"
        img = np.fliplr(img)
        if label.size:
            label[:, 1] = 1 - label[:, 1]
        return img, label

class YoloVFlipAug(AugBase):
    def __init__(self, deepvac_config):
        super(YoloVFlipAug, self).__init__(deepvac_config)

    def auditConfig(self):
        pass

    def __call__(self, img):
        img, label = self.auditInput(img, has_label=True)
        assert isinstance(label, np.ndarray) and label.ndim == 2, "label must be numpy.ndarray, and shape should be (n, 5)"
        img = np.flipud(img)
        if label.size:
            label[:, 2] = 1 - label[:, 2]
        return img, label

class YoloPerspectiveAug(AugBase):
    def __init__(self, deepvac_config):
        super(YoloPerspectiveAug, self).__init__(deepvac_config)
        self.border = deepvac_config.border

    def auditConfig(self):
        self.config.yolo_perspective_scale = self.addUserConfig('yolo_perspective_scale', self.config.yolo_perspective_scale, 0.5)
        self.config.yolo_perspective_shear = self.addUserConfig('yolo_perspective_shear', self.config.yolo_perspective_shear, 0.0)
        self.config.yolo_perspective_degrees = self.addUserConfig('yolo_perspective_degrees', self.config.yolo_perspective_degrees, 0.0)
        self.config.yolo_perspective_translate = self.addUserConfig('yolo_perspective_translate', self.config.yolo_perspective_translate, 0.1)
        self.config.yolo_perspective_perspective = self.addUserConfig('yolo_perspective_perspective', self.config.yolo_perspective_perspective, 0.0)

    def _box_candidates(self, box1, box2, wh_thr=2, ar_thr=20, area_thr=0.1):
        w1, h1 = box1[2] - box1[0], box1[3] - box1[1]
        w2, h2 = box2[2] - box2[0], box2[3] - box2[1]
        ar = np.maximum(w2 / (h2 + 1e-16), h2 / (w2 + 1e-16))
        return (w2 > wh_thr) & (h2 > wh_thr) & (w2 * h2 / (w1 * h1 + 1e-16) > area_thr) & (ar < ar_thr)

    def __call__(self, img):
        img, label = self.auditInput(img, has_label=True)
        assert isinstance(label, np.ndarray) and label.ndim == 2, "label must be numpy.ndarray, and shape should be (n, 5)"

        border = self.border
        h, w, c = img.shape

        width = int(w + border[1] * 2)
        height = int(h + border[0] * 2)
        # Center
        '''
            [[1, 0, -w/2],
             [0, 1, -h/2],
             [0, 0, 1.  ]]
        '''
        C = np.eye(3)
        C[0, 2] = -w / 2
        C[1, 2] = -h / 2
        # Perspective
        '''
            [[1, 0, 0],
             [0, 1, 0],
             [p, p, 1]]
        '''
        P = np.eye(3)
        P[2, 0] = random.uniform(-self.config.yolo_perspective_perspective, self.config.yolo_perspective_perspective)
        P[2, 1] = random.uniform(-self.config.yolo_perspective_perspective, self.config.yolo_perspective_perspective)
        # Rotation and Scale
        '''
            [[r, r, r],
             [r, r, r],
             [0, 0, 1]]
        '''
        R = np.eye(3)
        a = random.uniform(-self.config.yolo_perspective_degrees, self.config.yolo_perspective_degrees)
        s = random.uniform(1 - self.config.yolo_perspective_scale, 1 + self.config.yolo_perspective_scale)
        R[:2] = cv2.getRotationMatrix2D(angle=a, center=(0, 0), scale=s)
        # Shear
        '''
            [[1, s, 0],
             [s, 1, 0],
             [0, 0, 1]]
        '''
        S = np.eye(3)
        S[0, 1] = math.tan(random.uniform(-self.config.yolo_perspective_shear, self.config.yolo_perspective_shear) * math.pi / 180)
        S[1, 0] = math.tan(random.uniform(-self.config.yolo_perspective_shear, self.config.yolo_perspective_shear) * math.pi / 180)
        # Translation
        '''
            [[1, 0, t],
             [0, 1, t],
             [0, 0, 1]]
        '''
        T = np.eye(3)
        T[0, 2] = random.uniform(0.5 - self.config.yolo_perspective_translate, 0.5 + self.config.yolo_perspective_translate) * width
        T[1, 2] = random.uniform(0.5 - self.config.yolo_perspective_translate, 0.5 + self.config.yolo_perspective_translate) * height
        # Combined rotation matrix
        M = T @ S @ R @ P @ C
        # img augment and resize to img_size
        if (border[0] != 0) or (border[1] != 0) or (M != np.eye(3)).any():
            if self.config.yolo_perspective_perspective:
                img = cv2.warpPerspective(img, M, dsize=(width, height), borderValue=(114, 114, 114))
            else:
                img = cv2.warpAffine(img, M[:2], dsize=(width, height), borderValue=(114, 114, 114))

        n = len(label)
        if n:
            # warp points
            xy = np.ones((n * 4, 3))
            '''
                [[x1, y1],
                 [x2, y2],
                 [x1, y2],
                 [x2, y1]]
            '''
            xy[:, :2] = label[:, [1, 2, 3, 4, 1, 4, 3, 2]].reshape(n * 4, 2)
            xy = xy @ M.T
            if self.config.yolo_perspective_perspective:
                xy = (xy[:, :2] / xy[:, 2:3]).reshape(n, 8)
            else:
                xy = xy[:, :2].reshape(n, 8)
            # x: [[x1, x2, x1, x2], ...n...]
            x = xy[:, [0, 2, 4, 6]]
            # y: [[y1, y2, y2, y1], ...n...]
            y = xy[:, [1, 3, 5, 7]]
            # xy: [[xmin, ymin, xmax, ymax], ...n...]
            xy = np.concatenate((x.min(1), y.min(1), x.max(1), y.max(1))).reshape(4, n).T
            # clip boxes
            xy[:, [0, 2]] = xy[:, [0, 2]].clip(0, width)
            xy[:, [1, 3]] = xy[:, [1, 3]].clip(0, height)
            # filter candidates
            i = self._box_candidates(box1=label[:, 1:5].T * s, box2=xy.T)
            label = label[i]
            label[:, 1:5] = xy[i]
        return img, label


class YoloNormalizeAug(AugBase):
    def __init__(self, deepvac_config):
        super(YoloNormalizeAug, self).__init__(deepvac_config)

    def auditConfig(self):
        pass

    def __call__(self, img):
        '''
            1. [cls, x1, y1, x2, y2] -> [cls, cx, cy, w, h]
            2. cx, w normalized by img width, cy, h normalized by img height
        '''
        img, label = self.auditInput(img, has_label=True)
        assert isinstance(label, np.ndarray) and label.ndim == 2, "label must be numpy.ndarray, and shape should be (n, 5)"
        if not label.size:
            return img, label
        label[:, [3, 4]] -= label[:, [1, 2]]
        label[:, [1, 2]] += label[:, [3, 4]] / 2
        label[:, [1, 3]] /= img.shape[1]
        label[:, [2, 4]] /= img.shape[0]
        return img, label



import torch
import torch.nn as nn
from .conv_layer import Conv2dBN, Conv2dBNLeakyReLU

class SSH(nn.Module):
    def __init__(self, in_planes: int, out_planes: int):
        super(SSH, self).__init__()
        assert  out_planes % 4 == 0
        leaky = 0.1 if out_planes <= 64 else 0
        self.conv3X3 = Conv2dBN(in_planes, out_planes//2, padding=1)

        self.conv5X5_1 = Conv2dBNLeakyReLU(in_planes, out_planes//4, padding=1, leaky=leaky)
        self.conv5X5_2 = Conv2dBN(out_planes//4, out_planes//4, padding=1)

        self.conv7X7_2 = Conv2dBNLeakyReLU(out_planes//4, out_planes//4, padding=1, leaky=leaky)
        self.conv7x7_3 = Conv2dBN(out_planes//4, out_planes//4, padding=1)

        self.relu = nn.ReLU(inplace=True)

    def forward(self, input):
        conv3X3 = self.conv3X3(input)

        conv5X5_1 = self.conv5X5_1(input)
        conv5X5 = self.conv5X5_2(conv5X5_1)

        conv7X7_2 = self.conv7X7_2(conv5X5_1)
        conv7X7 = self.conv7x7_3(conv7X7_2)

        out = torch.cat([conv3X3, conv5X5, conv7X7], dim=1)
        out = self.relu(out)
        return out


from ..utils import LOG

class SyszuxFactory(object):
    def __init__(self):
        self.factory_dict = dict()
        self.locals = None
        self.auditConfig()
        self.initProducts()
        self.initMyProducts()

    def auditConfig(self):
        self.start_prefix = ''
        self.end_suffix = ''
        self.omit_list = []

    def initProducts(self):
        LOG.logE("You must reimplement initProducts() function.", exit=True)
    
    def initMyProducts(self):
        pass

    def addProduct(self, name, ins):
        LOG.logE("You must reimplement addProduct() function in subclass.", exit=True)

    def initSym(self):
        self.sym = dir(self.product_kind)

    def omitUnderScoreSym(self):
        self.sym = [x for x in self.sym if not x.startswith('_')]

    def selectStartPrefix(self):
        self.sym = [x for x in self.sym if x.startswith(self.start_prefix)]

    def selectEndSuffix(self):
        self.sym = [x for x in self.sym if x.endswith(self.end_suffix)]

    def omitOmitList(self):
        self.sym = [x for x in self.sym if x not in self.omit_list]

    def getSymsFromProduct(self):
        self.initSym()
        self.omitUnderScoreSym()
        self.selectStartPrefix()
        self.selectEndSuffix()
        self.omitOmitList()
        return self.sym

    def addProducts(self, product, product_name, start_prefix='', end_suffix='', omit_list=[]):
        if product_name not in self.locals:
            LOG.logE("{} not in locals(). You may forget set self.locals = locals in subclass auditConfig API.".format(product_name), exit=True)
            return
        self.product_kind = product
        self.omit_list = omit_list
        self.start_prefix = start_prefix
        self.end_suffix = end_suffix
        for sym in self.getSymsFromProduct():
            new_sym = sym
            if '_aug' not in product_name:
                new_sym = '{}.{}'.format(product_name, new_sym)
            self.addProduct(new_sym, '{}.{}'.format(product_name, sym) )

    def get(self, ins_name):
        try:
            x = self.factory_dict[ins_name]
        except:
            LOG.logE("ERROR! {} not found in factory.".format(ins_name), exit=True)
        return x


from ..utils import LOG
from .base import DeepvacCast

class TensorrtCast(DeepvacCast):
    def __init__(self, deepvac_train_config):
        super(TensorrtCast,self).__init__(deepvac_train_config)

    def auditConfig(self):
        if not self.config.trt_model_dir:
            return False
        
        if self.config.trt_enable_dynamic_input and self.config.onnx_dynamic_ax is None:
            LOG.logE("If you want to tensorrt support dynamic input, you must set onnx_dynamic_ax.", exit=True)
        
        return True

    def process(self, cast_output_file=None):
        try:
            import tensorrt as trt
        except:
            LOG.logE("You must install tensorrt package if you want to convert pytorch to onnx. 1. Download Tensorrt7.2.3(for CUDA11.0) from https://developer.nvidia.com/tensorrt \
            2. unpack Tensorrt*.tar.gz  3. pip install tensorrt-x-cpx-none-linux_x86_64.whl in Tensorrt*(your_tensorrt_path)/python", exit=True)
            return

        output_trt_file = self.config.trt_model_dir
        if cast_output_file:
            output_trt_file = '{}/trt__{}.trt'.format(self.config.output_dir, cast_output_file)
            self.config.trt_model_dir = output_trt_file
        
        LOG.logI("config.trt_model_dir found, save tensorrt model to {}...".format(self.config.trt_model_dir))

        #to onnx, also set self.config.onnx_model_dir, self.config.onnx_input_names and self.config.onnx_output_names
        self.exportOnnx()

        trt_logger = trt.Logger(trt.Logger.WARNING)
        with trt.Builder(trt_logger) as builder, builder.create_network(1 << (int)(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)) as network, trt.OnnxParser(network, trt_logger) as parser:
            builder.max_workspace_size = 4 << 30
            builder.max_batch_size = 1
            with open(self.config.onnx_model_dir, 'rb') as model:
                parser.parse(model.read())
            config = builder.create_builder_config()
            if self.config.trt_enable_dynamic_input:
                profile = builder.create_optimization_profile()
                profile.set_shape(self.config.onnx_input_names[0], self.config.trt_input_min_dims, self.config.trt_input_opt_dims, self.config.trt_input_max_dims)
                config.add_optimization_profile(profile)
            engine = builder.build_engine(network, config)
            with open(output_trt_file, "wb") as f:
                f.write(engine.serialize())
        LOG.logI("Pytorch model convert to TensorRT model succeed, save model in {}".format(output_trt_file))


# -*- coding: utf-8 -*-
import os
import sys
import numpy as np
from numpy.linalg import inv, norm, lstsq
from numpy.linalg import matrix_rank as rank
import time
import cv2

class AlignFace(object):
    def __init__(self):
        # reference facial points, a list of coordinates (x,y)
        self.REFERENCE_FACIAL_POINTS_96x112 = [
            [30.29459953, 51.69630051],
            [65.53179932, 51.50139999],
            [48.02519989, 71.73660278],
            [33.54930115, 92.3655014],
            [62.72990036, 92.20410156]
        ]
        # made by gemfield
        self.REFERENCE_FACIAL_POINTS_112x112 = [
            [38.29459953, 51.69630051],
            [73.53179932, 51.50139999],
            [56.02519989, 71.73660278],
            [41.54930115, 92.3655014 ],
            [70.72990036, 92.20410156]
        ]

    def __call__(self, frame, facial_5pts):
        # shape from (10,) to (2, 5)
        # x1,x2...y1,y2... or  x1,y1,x2,y2...
        facial = []
        x = facial_5pts[::2]
        y = facial_5pts[1::2]
        facial.append(x)
        facial.append(y)
        dst_img = self.warpAndCrop(frame, facial, (112, 112))
        return dst_img

    def warpAndCrop(self, src_img, facial_pts, crop_size):
        reference_pts = self.REFERENCE_FACIAL_POINTS_112x112
        ref_pts = np.float32(reference_pts)
        ref_pts_shp = ref_pts.shape

        if ref_pts_shp[0] == 2:
            ref_pts = ref_pts.T

        src_pts = np.float32(facial_pts)
        src_pts_shp = src_pts.shape
        if max(src_pts_shp) < 3 or min(src_pts_shp) != 2:
            raise Exception('facial_pts.shape must be (K,2) or (2,K) and K>2')
        # 2*5 to 5*2
        if src_pts_shp[0] == 2:
            src_pts = src_pts.T

        if src_pts.shape != ref_pts.shape:
            raise Exception('facial_pts and reference_pts must have the same shape: {} vs {}'.format(src_pts.shape, ref_pts.shape) )

        tfm = self.getAffineTransform(src_pts, ref_pts)

        face_img = cv2.warpAffine(src_img, tfm, (crop_size[0], crop_size[1]))
        return face_img

    def getAffineTransform(self, uv, xy):
        options = {'K': 2}
        # Solve for trans1
        trans1, trans1_inv = self.findNonreflectiveSimilarity(uv, xy, options)
        # manually reflect the xy data across the Y-axis
        xyR = xy
        xyR[:, 0] = -1 * xyR[:, 0]

        trans2r, trans2r_inv = self.findNonreflectiveSimilarity(uv, xyR, options)

        # manually reflect the tform to undo the reflection done on xyR
        TreflectY = np.array([
            [-1, 0, 0],
            [0, 1, 0],
            [0, 0, 1]
        ])

        trans2 = np.dot(trans2r, TreflectY)

        # Figure out if trans1 or trans2 is better
        xy1 = self.tformfwd(trans1, uv)
        norm1 = norm(xy1 - xy)

        xy2 = self.tformfwd(trans2, uv)
        norm2 = norm(xy2 - xy)

        if norm1 <= norm2:
            trans = trans1
        else:
            trans2_inv = inv(trans2)
            trans = trans2

        cv2_trans = trans[:, 0:2].T
        return cv2_trans

    def findNonreflectiveSimilarity(self, uv, xy, options=None):
        options = {'K': 2}

        K = options['K']
        M = xy.shape[0]
        x = xy[:, 0].reshape((-1, 1))
        y = xy[:, 1].reshape((-1, 1))

        tmp1 = np.hstack((x, y, np.ones((M, 1)), np.zeros((M, 1))))
        tmp2 = np.hstack((y, -x, np.zeros((M, 1)), np.ones((M, 1))))
        X = np.vstack((tmp1, tmp2))

        u = uv[:, 0].reshape((-1, 1))
        v = uv[:, 1].reshape((-1, 1))
        U = np.vstack((u, v))

        if rank(X) >= 2 * K:
            r, _, _, _ = lstsq(X, U, rcond=-1)
            r = np.squeeze(r)
        else:
            raise Exception('cp2tform:twoUniquePointsReq')
        sc = r[0]
        ss = r[1]
        tx = r[2]
        ty = r[3]

        Tinv = np.array([
            [sc, -ss, 0],
            [ss, sc, 0],
            [tx, ty, 1]
        ])

        T = inv(Tinv)
        T[:, 2] = np.array([0, 0, 1])
        return T, Tinv

    def tformfwd(self, trans, uv):
        uv = np.hstack((
            uv, np.ones((uv.shape[0], 1))
        ))
        xy = np.dot(uv, trans)
        xy = xy[:, 0:-1]
        return xy

    def drawBoxes(self, img, imgname):
        write_path = './results/{}/bbox/'.format( time.strftime("%Y%m%d", time.localtime()) )
        if not os.path.exists(write_path):
            os.makedirs(write_path)
        cv2.imwrite('{}{}_aligned.jpg'.format(write_path, imgname), img)

class AlignFaceWith51Points(AlignFace):
    def __init__(self):
        super(AlignFaceWith51Points, self).__init__()
        self.REFERENCE_FACIAL_POINTS_112x112 = self.getPosition()

    def getPosition(self):

        x = [0.000213256, 0.0752622, 0.18113, 0.29077, 0.393397, 0.586856, 0.689483, 0.799124,
                    0.904991, 0.98004, 0.490127, 0.490127, 0.490127, 0.490127, 0.36688, 0.426036,
                    0.490127, 0.554217, 0.613373, 0.121737, 0.187122, 0.265825, 0.334606, 0.260918,
                    0.182743, 0.645647, 0.714428, 0.793132, 0.858516, 0.79751, 0.719335, 0.254149,
                    0.340985, 0.428858, 0.490127, 0.551395, 0.639268, 0.726104, 0.642159, 0.556721,
                    0.490127, 0.423532, 0.338094, 0.290379, 0.428096, 0.490127, 0.552157, 0.689874,
                    0.553364, 0.490127, 0.42689]

        y = [0.106454, 0.038915, 0.0187482, 0.0344891, 0.0773906, 0.0773906, 0.0344891,
                    0.0187482, 0.038915, 0.106454, 0.203352, 0.307009, 0.409805, 0.515625, 0.587326,
                    0.609345, 0.628106, 0.609345, 0.587326, 0.216423, 0.178758, 0.179852, 0.231733,
                    0.245099, 0.244077, 0.231733, 0.179852, 0.178758, 0.216423, 0.244077, 0.245099,
                    0.780233, 0.745405, 0.727388, 0.742578, 0.727388, 0.745405, 0.780233, 0.864805,
                    0.902192, 0.909281, 0.902192, 0.864805, 0.784792, 0.778746, 0.785343, 0.778746,
                    0.784792, 0.824182, 0.831803, 0.824182]

        x, y = np.array(x), np.array(y)
        x = x * 112
        y = y * 112
        return np.array(list(zip(x, y)))

    def getAffineTransform(self, src_pts, ref_pts):
        src_pts = np.matrix(src_pts)
        ref_pts = np.matrix(ref_pts)
        mean1 = np.mean(src_pts, axis=0)
        mean2 = np.mean(ref_pts, axis=0)
        src_pts -= mean1
        ref_pts -= mean2
        std1 = np.std(src_pts)
        std2 = np.std(ref_pts)
        src_pts /= std1
        ref_pts /= std2

        U, S, Vt = np.linalg.svd(src_pts.T * ref_pts)
        R = (U * Vt).T
        M = np.vstack([
            np.hstack(((std2 / std1) * R,
            mean2.T - (std2 / std1) * R * mean1.T)),
            np.matrix([0., 0., 1.])
        ])

        return M[:2]



import subprocess
import tempfile
import copy
import torch
from torch.quantization import quantize_dynamic_jit, per_channel_dynamic_qconfig
from torch.quantization import get_default_qconfig, quantize_jit
from torch.quantization import default_dynamic_qconfig, float_qparams_weight_only_qconfig
from torch.quantization.quantize_fx import prepare_fx, convert_fx, prepare_qat_fx
import torch.nn as nn
from ..utils import LOG

class DeepvacCast(object):
    def __init__(self, deepvac_train_config):
        self.config = deepvac_train_config
        self.proceed = False
        self.auditFinalConfig()

    def auditConfig(self):
        LOG.logW("You should reimplement auditConfig in your subclass.")

    def auditFinalConfig(self):
        if not self.auditConfig():
            return
        if self.config.net is None:
            LOG.logE("You must set config.train.net in config.py", exit=True)

        if self.config.sample is None:
            LOG.logE("You must set config.train.sample, in general, this is done be Deepvac Framework.", exit=True)
        
        self.net = copy.deepcopy(self.config.ema if self.config.ema else self.config.net)
        if self.config.cast2cpu:
            self.net.cpu()
            self.config.sample = self.config.sample.to('cpu')
        
        self.proceed = True
        
    def process(self, cast_output_file=None):
        LOG.logE("You must reimplement process() in your subclass.", exit=True)

    def exportOnnx(self):
        if self.config.onnx_input_names is None:
            self.config.onnx_input_names = ["input"]
        
        if self.config.onnx_output_names is None:
            self.config.onnx_output_names = ["output"]

        if not self.config.onnx_model_dir:
            f = tempfile.NamedTemporaryFile(delete=False)
            self.config.onnx_model_dir = f.name

        torch.onnx.export(self.net, self.config.sample, self.config.onnx_model_dir, 
            input_names=self.config.onnx_input_names, output_names=self.config.onnx_output_names, 
            dynamic_axes=self.config.onnx_dynamic_ax, opset_version=self.config.onnx_version, export_params=True)
        LOG.logI("Pytorch model convert to ONNX model succeed, save model in {}".format(self.config.onnx_model_dir))

    def __call__(self, cast_output_file=None):
        if not self.proceed:
            return
        self.process(cast_output_file)

def calibrate(model, data_loader):
    model.eval()
    with torch.no_grad():
        for sample, target in data_loader:
            model(sample)
            
class ScriptModel(object):
    def __init__(self, deepvac_train_config, output_file, backend = 'fbgemm'):
        self.config = deepvac_train_config
        self.input_net = copy.deepcopy(self.config.ema if self.config.ema else self.config.net)
        self.input_net.to(self.config.sample.device)
        self.input_net.eval()
        self.output_file = output_file
        self.backend = backend
        self.dq_output_file = '{}.dq'.format(output_file)
        self.sq_output_file = '{}.sq'.format(output_file)
        self.d_qconfig_dict = {'': per_channel_dynamic_qconfig}
        self.s_qconfig_dict = {'': get_default_qconfig(self.backend) }

    def _freeze_jit(self, model):
        LOG.logE("You must reimplement _freeze_jit() in subclass.", exit=True)
    
    def _jit(self, model):
        LOG.logE("You must reimplement _jit() in subclass.", exit=True)

    def getCalibrateLoader(self):
        loader = self.config.test_loader if self.config.is_forward_only else self.config.val_loader
        if loader is None:
            LOG.logE("You enabled config.static_quantize_dir, but didn't provide config.test_loader in forward-only mode, or self.val_loader in train mode.", exit=True)
        return loader

    def save(self):
        with torch.no_grad():
            torch.jit.save(self._freeze_jit(self.input_net), self.output_file)

    def saveDQ(self):
        LOG.logI("Pytorch model dynamic quantize starting, will save model in {}".format(self.dq_output_file))
        with torch.no_grad():
            quantized_model = quantize_dynamic_jit(self._jit(), self.d_qconfig_dict)
            torch.jit.save(quantized_model, self.dq_output_file)
        LOG.logI("Pytorch model dynamic quantize succeeded, saved model in {}".format(self.dq_output_file))

    def saveSQ(self):
        loader = self.getCalibrateLoader()
        LOG.logI("Pytorch model static quantize starting, will save model in {}".format(self.sq_output_file))
        with torch.no_grad():
            quantized_model = quantize_jit(self._jit(), self.s_qconfig_dict, calibrate, [loader], inplace=False, debug=False)
            torch.jit.save(quantized_model, self.sq_output_file)
        LOG.logI("Pytorch model static quantize succeeded, saved model in {}".format(self.sq_output_file))

class SaveModelByTrace(ScriptModel):
    def _freeze_jit(self, model):
        return torch.jit.freeze(torch.jit.trace(model, self.config.sample).eval() )
    
    def _jit(self, model):
        return torch.jit.trace(model, self.config.sample).eval()

class SaveModelByScript(ScriptModel):
    def _freeze_jit(self, model):
        return torch.jit.freeze(torch.jit.script(model).eval() )

    def _jit(self, model):
        return torch.jit.script(model).eval()

class FXQuantize(ScriptModel):
    def saveDQ(self):
        qconfig_dict = {
            "object_type": [
                (nn.Embedding, float_qparams_weight_only_qconfig),
                (nn.LSTM, default_dynamic_qconfig),
                (nn.Linear, default_dynamic_qconfig)
            ]
        }
        prepared_model = prepare_fx(self.input_net, qconfig_dict)
        quantized_model = convert_fx(prepared_model)
        torch.jit.save(self._freeze_jit(quantized_model), self.dq_output_file)

    def saveSQ(self):
        loader = self.getCalibrateLoader()
        prepared_model = prepare_fx(self.input_net, self.s_qconfig_dict)
        #print(prepared_model.graph)
        calibrate(prepared_model, loader)
        quantized_model = convert_fx(prepared_model)
        torch.jit.save(self._freeze_jit(quantized_model), self.sq_output_file)

class EagerQuantize(ScriptModel):
    def saveDQ(self):
        quantized_model = torch.quantization.quantize_dynamic(self.input_net, inplace=False)
        torch.jit.save(self._freeze_jit(quantized_model), self.dq_output_file)

    def saveSQ(self):
        loader = self.getCalibrateLoader()
        fused_model = auto_fuse_model(self.input_net)
        fused_model.qconfig = torch.quantization.get_default_qconfig(self.backend)
        prepared_model = torch.quantization.prepare(fused_model, inplace=False)
        calibrate(prepared_model, loader)
        quantized_model = torch.quantization.convert(prepared_model, inplace=False)
        torch.jit.save(self._freeze_jit(quantized_model), self.sq_output_file)

class FXQuantizeAndScript(FXQuantize, SaveModelByScript):
    pass

class FXQuantizeAndTrace(FXQuantize, SaveModelByTrace):
    pass

class EagerQuantizeAndScript(EagerQuantize, SaveModelByScript):
    pass

class EagerQuantizeAndTrace(EagerQuantize, SaveModelByTrace):
    pass


import sys
import time
import numpy as np
import cv2
from PIL import Image
import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision import transforms as trans
from torchvision.models import resnet50
from ..core.config import config
from ..core.deepvac import Deepvac, DeepvacTrain
from ..datasets.os_walk import OsWalkDataset
from ..datasets.file_line import FileLineDataset
from ..utils import LOG
from .weights_init import initWeightsKaiming
from .bottleneck_layer import Bottleneck
from .conv_layer import Conv2dBNReLU

class ResnetBasicBlock(nn.Module):
    expansion: int = 1
    def __init__(self, inplanes: int, outplanes: int, stride: int = 1):
        super(ResnetBasicBlock, self).__init__()
        self.conv1 = Conv2dBNReLU(in_planes=inplanes, out_planes=outplanes, kernel_size=3, stride=stride)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(outplanes, outplanes, kernel_size=3, padding=1)
        self.bn2 = nn.BatchNorm2d(outplanes)
        self.downsample = None
        if stride != 1 or inplanes != outplanes:
            self.downsample = nn.Sequential(nn.Conv2d(inplanes, outplanes, kernel_size=1, stride=stride, bias=False), nn.BatchNorm2d(outplanes))

    def forward(self, x):
        identity = x
        out = self.conv1(x)
        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)
        out += identity
        return self.relu(out)

class ResNet18(nn.Module):
    def __init__(self, class_num: int = 1000):
        super(ResNet18, self).__init__()
        self.inplanes = 64
        self.class_num = class_num
        self.auditConfig()
        self.conv1 = Conv2dBNReLU(in_planes=3, out_planes=self.inplanes, kernel_size=7, stride=2)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        layers = []
        #init the 4 layers
        for outp, layer_num, stride in self.cfgs:
            layers.append(self.block(self.inplanes, outp, stride))
            self.inplanes = outp * self.block.expansion
            for _ in range(1, layer_num):
                layers.append(self.block(self.inplanes, outp))

        self.layer = nn.Sequential(*layers)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

        self.initFc()
        initWeightsKaiming(self)

    def forward(self, x):
        x = self.conv1(x)
        x = self.maxpool(x)
        x = self.layer(x)
        x = self.avgpool(x)
        return self.forward_cls(x)

    def forward_cls(self, x):
        x = torch.flatten(x, 1)
        x = self.fc(x)
        return x

    def initFc(self):
        self.fc = nn.Linear(512 * self.block.expansion, self.class_num)

    def auditConfig(self):
        self.block = ResnetBasicBlock
        self.cfgs = [
            # outp, layer_num, s
            [64,   2,  1],
            [128,  2,  2],
            [256,  2,  2],
            [512,  2,  2]
        ]

class ResNet34(ResNet18):
    def __init__(self,class_num: int = 1000):
        super(ResNet34, self).__init__(class_num)

    def auditConfig(self):
        self.block = ResnetBasicBlock
        self.cfgs = [
            # outp, layer_num, s
            [64,   3,  1],
            [128,  4,  2],
            [256,  6,  2],
            [512,  3,  2]
        ]
    
class ResNet50(ResNet18):
    def __init__(self,class_num: int = 1000):
        super(ResNet50, self).__init__(class_num)

    def auditConfig(self):
        self.block = Bottleneck
        self.cfgs = [
            # outp, layer_num, s
            [64,   3,  1],
            [128,  4,  2],
            [256,  6,  2],
            [512,  3,  2]
        ]

class ResNet101(ResNet18):
    def __init__(self,class_num: int = 1000):
        super(ResNet101, self).__init__(class_num)

    def auditConfig(self):
        self.block = Bottleneck
        self.cfgs = [
            # outp, layer_num, s
            [64,   3,  1],
            [128,  4,  2],
            [256,  23,  2],
            [512,  3,  2]
        ]

class ResNet152(ResNet18):
    def __init__(self,class_num: int = 1000):
        super(ResNet152, self).__init__(class_num)

    def auditConfig(self):
        self.block = Bottleneck
        self.cfgs = [
            # outp, layer_num, s
            [64,   3,  1],
            [128,  8,  2],
            [256,  36,  2],
            [512,  3,  2]
        ]

class ResNet18OCR(ResNet18):
    def __init__(self):
        super(ResNet18OCR, self).__init__()

    def auditConfig(self):
        self.block = ResnetBasicBlock
        self.cfgs = [
            [64,   2,  1],
            [128,  2,  (2,1)],
            [256,  2,  (2,1)],
            [512,  2,  (2,1)]
        ]

    def initFc(self):
        self.avgpool = nn.AvgPool2d((2,2))
    
    def forward(self, x):
        x = self.conv1(x)
        x = self.layer(x)
        x = self.avgpool(x)
        b, c, h, w = x.size()
        assert h == 1, "the height of conv must be 1"
        x = x.squeeze(2) # b *512 * width
        x = x.permute(2, 0, 1)  # [w, b, c]
        return x


class ResNet50Train(DeepvacTrain):
    def initNetWithCode(self):
        self.net = ResNet50()

    def initTrainLoader(self):
        self.train_dataset = FileLineDataset(self.conf.train)
        self.train_loader = torch.utils.data.DataLoader(self.train_dataset, batch_size=self.conf.train.batch_size, pin_memory=False)

    def initValLoader(self):
        self.val_dataset = FileLineDataset(self.conf.val)
        self.val_loader = torch.utils.data.DataLoader(self.val_dataset, batch_size=1, pin_memory=False)

    def postEpoch(self):
        if self.is_train:
            return
        self.accuracy = 1
        LOG.logI('Test accuray: {:.4f}'.format(self.accuracy))

class ResnetClsTestDataset(OsWalkDataset):
    def __init__(self, deepvac_config):
        super(ResnetClsTestDataset, self).__init__(deepvac_config)

    def __getitem__(self, index):
        filepath = self.files[index]
        sample = Image.open(filepath).convert('RGB')
        if self.transform is not None:
            sample = self.transform(sample)
        return sample, filepath

class ResNet50Test(Deepvac):
    def __init__(self, deepvac_config):
        super(ResNet50Test, self).__init__(deepvac_config)

    def initNetWithCode(self):
        self.net = resnet50()
    
    def initTestLoader(self):
        if self.conf.disable_loader:
            return
        self.test_dataset = ResnetClsTestDataset(self.conf)
        self.test_loader = torch.utils.data.DataLoader(self.test_dataset, batch_size=1, pin_memory=False)

    def validate(self, t, img_path):
        LOG.logI('---------------VALIDATE BEGIN---------------')
        img = cv2.imread(img_path)
        if img.shape is None:
            LOG.logE('illegal image detected in validate!',exit=True)

        img = cv2.resize(img, (t.size(4), t.size(3)))
        transformer = trans.Compose([
            trans.ToTensor(),
            trans.Normalize([0.5, 0.5, 0.5], [0.5, 0.5, 0.5])
        ])

        input_t = transformer(img)
        input_t = input_t.to(self.device)

        torch.cuda.synchronize()
        start = time.time()

        preds = self.net(input_t.unsqueeze(0))
        torch.cuda.synchronize()
        end = time.time()
 
        softmaxs = F.softmax( preds, dim=1 )
        max_res = torch.max(softmaxs, dim=1)
        max_probability, max_index = max_res

        LOG.logI('Overall process time in validate: {} | Index: {} | Class: {} | Probability: {}'.format(end-start, max_index, max_index, max_probability))
        LOG.logI('---------------VALIDATE END---------------\n')

    def warmUp(self, t):
        LOG.logI('---------------WARMUP BEGIN---------------')

        for i in range(10):
            ti = torch.rand(( t.size(1), t.size(2), t.size(3), t.size(4) ), dtype = torch.float).to(self.device)
 
            torch.cuda.synchronize()
            start = time.time()

            resnet_out = self.net(ti)

            torch.cuda.synchronize()
            end = time.time()
        
            LOG.logI('Overall process time in warmup: {}'.format(end-start))

        LOG.logI('---------------WARMUP END---------------')

    def benchmark(self, t, img_path):
        self.validate(t, img_path)
        self.warmUp(t)

        item_num = t.size(0)
        LOG.logI('---------------BENCHMARK BEGIN--------------- {}'.format(item_num))

        torch.cuda.synchronize()
        start = time.time()

        for i in range(item_num):
            ti = t[i]
            torch.cuda.synchronize()
            tick = time.time()
            resnet_out = self.net(ti)
            
            torch.cuda.synchronize()
            tock = time.time()
            LOG.logI('forward:model forward time: {}'.format(tock-tick))

        torch.cuda.synchronize()
        end = time.time()

        print("|Model|Engine|Input size|forward time|")
        print("|-----|-------|----------|-----------|")
        print("|Resnet50|libtorch|{}x{}|{}|".format(t.size(4), t.size(3), (end-start)/item_num))
        LOG.logI('---------------BENCHMARK END---------------')

    def process(self):
        for input_tensor, path in self.test_loader:
            preds = self.net(input_tensor.to(self.device))
            softmaxs = F.softmax( preds, dim=1 )
            max_res = torch.max(softmaxs, dim=1)
            max_probability, max_index = max_res
            LOG.logI("path: {}, max_probability:{}, max_index:{}".format(path[0], max_probability.item(), max_index.item()))

def auditConfig():
    config.test.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    config.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    config.test.disable_git = True
    config.test.script_model_dir = "./gemfield_script.pt"
    config.test.transform = trans.Compose([
        trans.Resize((224, 224)),
        trans.ToTensor(),
        trans.Normalize([0.5, 0.5, 0.5], [0.5, 0.5, 0.5])
    ])

    config.disable_git = True
    #train stuff
    config.momentum = 0.9
    config.weight_decay = 5e-4
    config.gamma = 0.1
    config.lr = 1e-3

    config.epoch_num = 100
    config.save_num = 1
    config.log_every = 100
    config.num_workers = 4
    config.lr_step = [50, 70, 90]
    config.lr_factor = 0.1

    config.train.shuffle = True
    config.train.batch_size = 1

    config.train.transform = trans.Compose([
        trans.ToTensor(),
        trans.Normalize([0.5, 0.5, 0.5], [0.5, 0.5, 0.5])
    ])
    config.val.transform = trans.Compose([
        trans.ToTensor(),
        trans.Normalize([0.5, 0.5, 0.5], [0.5, 0.5, 0.5])
    ])

if __name__ == "__main__":
    if(len(sys.argv) < 2):
        LOG.logE("Usage: python -m deepvac.syszux_resnet <train|test|benchmark> <pretrained_model.pth> <your_input>", exit=True)
    
    op = sys.argv[1]
    if op not in ('train','test','benchmark'):
        LOG.logE("Usage: python -m deepvac.syszux_resnet <train|test|benchmark> <pretrained_model.pth> <your_input>", exit=True)
    #
    auditConfig()

    if op == 'train':
        if(len(sys.argv) != 6):
            LOG.logE("Usage: python -m deepvac.syszux_resnet train <pretrained_model.pth> <train_val_data_dir_prefix> <train.txt> <val.txt>", exit=True)

        config.model_path = sys.argv[2]
        config.train.fileline_data_path_prefix = sys.argv[3]
        config.train.fileline_path = sys.argv[4]
        config.val.fileline_data_path_prefix = sys.argv[3]
        config.val.fileline_path = sys.argv[5]
        train = ResNet50Train(config)
        train()

    if op == 'test':
        if(len(sys.argv) != 4):
            LOG.logE("Usage: python -m deepvac.syszux_resnet test <pretrained_model.pth> <your_test_img_input_dir>", exit=True)

        config.test.model_path = sys.argv[2]
        config.test.static_quantize_dir = "./static_quantize.pt"
        config.test.input_dir = sys.argv[3]
        test = ResNet50Test(config.test)
        test()

    if op == 'benchmark':
        if(len(sys.argv) != 4):
            LOG.logE("Usage: python -m deepvac.syszux_resnet benchmark <pretrained_model.pth> <your_input_img.jpg>", exit=True)

        config.test.model_path = sys.argv[2]
        config.test.disable_loader = True
        img_path = sys.argv[3]

        t224x224 = torch.rand((100, 1, 3, 224, 224), dtype = torch.float).to(config.test.device)
        t640x640 = torch.rand((100, 1, 3, 640, 640), dtype = torch.float).to(config.test.device)
        t1280x720 = torch.rand((50, 1, 3, 720, 1280), dtype = torch.float).to(config.test.device)
        t1280x1280 = torch.rand((50, 1, 3, 1280, 1280), dtype = torch.float).to(config.test.device)

        test = ResNet50Test(config.test)
        test.benchmark(t224x224, img_path)
        test.benchmark(t640x640, img_path)
        test.benchmark(t1280x720, img_path)
        test.benchmark(t1280x1280, img_path)


from typing import List
import torch
import torch.nn as nn

class Concat(nn.Module):
    # Concatenate a list of tensors along dimension
    def __init__(self, dimension=1):
        super(Concat, self).__init__()
        self.d = dimension

    def forward(self, x: List[torch.Tensor]) -> torch.Tensor:
        return torch.cat(x, self.d)

class Focus(nn.Module):
    # Focus wh information into c-space
    def __init__(self, in_planes, out_planes, kernel_size=1, stride=1, eps=1e-5, momentum=0.1, padding=None, groups=1):
        super(Focus, self).__init__()
        self.conv = Conv2dBNHardswish(in_planes * 4, out_planes, kernel_size, stride, eps, momentum, padding, groups)

    def forward(self, x):  # x(b,c,w,h) -> y(b,4c,w/2,h/2)
        return self.conv(torch.cat([x[..., ::2, ::2], x[..., 1::2, ::2], x[..., ::2, 1::2], x[..., 1::2, 1::2]], 1))


import sys
import os
from datetime import datetime
from enum import Enum
import logging
import subprocess

def getCurrentGitBranch():
    try:
        branch = subprocess.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD']).strip().decode()
    except:
        branch = None
    return branch

def getTime():
    return (str(datetime.now())[:-10]).replace(' ','-').replace(':','-')

def getArgv():
    argv = ''.join(sys.argv)
    argv = argv.replace(' ','').replace('/','').replace('-','_').replace('.py','')
    return argv[:64]

deepvac_branch_name = getCurrentGitBranch()
if deepvac_branch_name is None:
    deepvac_branch_name = 'not_in_git'

deepvac_pid = os.getpid()
deepvac_time = getTime()

#according deepvac standard, log should in log directory.
if not os.path.exists('log'):
    os.makedirs('log')

logger=logging.getLogger("DEEPVAC")
logger.setLevel(logging.DEBUG)

deepvac_log_format = '%(asctime)s %(levelname)-8s %(message)s'
fh = logging.FileHandler('log/{}:{}:{}:{}.log'.format(deepvac_pid, getArgv(),deepvac_time, deepvac_branch_name))
fh.setLevel(logging.INFO)
fh.setFormatter( logging.Formatter(deepvac_log_format) )
logger.addHandler(fh)

# add console output
console_log_format = '%(asctime)s {}:{} %(levelname)-8s %(message)s'.format(deepvac_pid, deepvac_branch_name)
console = logging.StreamHandler()
console.setLevel(logging.INFO)
console.setFormatter( logging.Formatter(console_log_format) )
logger.addHandler(console)

logger.info("deepvac log imported...")

class LOG(object):
    class S(Enum):
        I = 'Info'
        W = 'Warning'
        E = 'Error'

    LOG_LEVEL = S.I
    logfunc = {S.I: lambda x : logger.info(x),
                S.W: lambda x : logger.warning(x),
                S.E: lambda x : logger.error(x)
                }
            
    intfunc = {S.I: 0, S.W: 1, S.E: 2}
    @staticmethod
    def log(level, str):
        if level not in LOG.logfunc.keys():
            LOG.logfunc[LOG.S.E]("incorrect value of parameter level when call log function.")

        if LOG.intfunc[level] < LOG.intfunc[LOG.LOG_LEVEL]:
            return 

        LOG.logfunc[level](str)

    @staticmethod
    def logI(str):
        LOG.log(LOG.S.I, str)

    @staticmethod
    def logW(str):
        LOG.log(LOG.S.W, str)

    @staticmethod
    def logE(str, exit=False):
        LOG.log(LOG.S.E, str)
        if exit:
            sys.exit(1)

# according to DeepVAC Standard, we must run in git repo
def assertAndGetGitBranch(is_disable_git):
    if os.environ.get("disable_git"):
        branch = "disable_git"
        return branch

    if is_disable_git:
        branch = "disable_git"
        return branch

    branch = getCurrentGitBranch()
    if branch is None:
        LOG.logE('According to deepvac standard, you must working in a git repo.', exit=True)

    if len(branch) < 4:
        LOG.logE('According to deepvac standard, your git branch name is too short: {}'.format(branch), exit=True)

    LOG.logI('You are running on git branch: {}'.format(branch))

    if branch.startswith('LTS_'):
        return branch

    if branch.startswith('PROTO_'):
        return branch

    if branch in ['master','main']:
        LOG.logW('However, DeepVAC does not suggest running on the master/main branch.')
        return branch

    LOG.logE('According to DeepVAC standard, git branch name should start from LTS_ or PROTO_: {}'.format(branch), exit=True)



from .trace import TraceCast
from .script import ScriptCast
from .onnx import OnnxCast
from .coreml import CoremlCast
from .ncnn import NcnnCast
from .tensorrt import TensorrtCast

caster_list = ['Trace', 'Script', 'Onnx', 'Coreml', 'Ncnn', 'Tensorrt']
def export3rd(deepvac_train_config, output_file=None):
    for caster_name in caster_list:
        caster = eval('{}Cast(deepvac_train_config)'.format(caster_name))
        caster(output_file)
        



import cv2
import numpy as np
import random
from .base_aug import AugBase

class ImageWithMasksRandomHorizontalFlipAug(AugBase):
    def __init__(self, deepvac_config):
        super(ImageWithMasksRandomHorizontalFlipAug, self).__init__(deepvac_config)

    def auditConfig(self):
        pass

    def __call__(self, imgs):
        img, label = self.auditInput(imgs, has_label=True)
        imgs = [img]
        imgs.extend(label)
        for i in range(len(imgs)):
            imgs[i] = np.flip(imgs[i], axis=1)
        return [imgs[0],imgs[1:]]

class ImageWithMasksRandomRotateAug(AugBase):
    def __init__(self, deepvac_config):
        super(ImageWithMasksRandomRotateAug, self).__init__(deepvac_config)

    def auditConfig(self):
        self.config.image_with_masks_random_rotate_max_angle = self.addUserConfig('image_with_masks_random_rotate_max_angle', self.config.image_with_masks_random_rotate_max_angle, 10)

    def __call__(self, imgs):
        img, label = self.auditInput(imgs, has_label=True)
        imgs = [img]
        imgs.extend(label)
        angle = random.random() * 2 * self.config.image_with_masks_random_rotate_max_angle - self.config.image_with_masks_random_rotate_max_angle
        for i in range(len(imgs)):
            img = imgs[i]
            w, h = img.shape[:2]
            rotation_matrix = cv2.getRotationMatrix2D((h / 2, w / 2), angle, 1)
            img_rotation = cv2.warpAffine(img, rotation_matrix, (h, w))
            imgs[i] = img_rotation
        return [imgs[0],imgs[1:]]

class ImageWithMasksRandomCropAug(AugBase):
    def __init__(self, deepvac_config):
        super(ImageWithMasksRandomCropAug, self).__init__(deepvac_config)
        self.auditUserConfig("img_size")

    def auditConfig(self):
        self.config.image_with_masks_random_crop_p = self.addUserConfig('image_with_masks_random_crop_p', self.config.image_with_masks_random_crop_p, 3.0 / 8.0)

    def __call__(self, imgs):
        img, label = self.auditInput(imgs, has_label=True)
        h, w, _ = img.shape

        imgs = [img]
        imgs.extend(label)
        if max(h, w) <= self.config.img_size:
            return [imgs[0],imgs[1:]]

        img_size = (self.config.img_size,) * 2
        th = min(h, img_size[0])
        tw = min(w, img_size[0])

        if random.random() > self.config.image_with_masks_random_crop_p and np.max(imgs[1]) > 0:
            tl = np.min(np.where(imgs[1] > 0), axis = 1) - img_size
            tl[tl < 0] = 0
            br = np.max(np.where(imgs[1] > 0), axis = 1) - img_size
            br[br < 0] = 0
            br[0] = min(br[0], h - th)
            br[1] = min(br[1], w - tw)

            i = random.randint(tl[0], br[0])
            j = random.randint(tl[1], br[1])
        else:
            i = random.randint(0, h - th)
            j = random.randint(0, w - tw)

        # return i, j, th, tw
        for idx in range(len(imgs)):
            if len(imgs[idx].shape) == 3:
                imgs[idx] = imgs[idx][i:i + th, j:j + tw, :]
            else:
                imgs[idx] = imgs[idx][i:i + th, j:j + tw]
        return [imgs[0],imgs[1:]]




import torch.nn as nn
from .module_helper import makeDivisible
from .activate_layer import hswish
from .conv_layer import Conv2dBNHswish
from .res_layer import InvertedResidual
from .weights_init import initWeightsKaiming

class MobileNetV3(nn.Module):
    def __init__(self, class_num=1000, width_mult=1.):
        super(MobileNetV3, self).__init__()
        self.width_mult = width_mult
        self.class_num = class_num
        # setting of inverted residual blocks
        self.auditConfig()
        # building first layer
        input_channel = makeDivisible(16 * self.width_mult, 8)
        layers = [Conv2dBNHswish(3, input_channel, stride=2)]

        # building inverted residual blocks
        for k, t, c, use_se, use_hs, s in self.cfgs:
            exp_size = makeDivisible(input_channel * t, 8)
            output_channel = makeDivisible(c * self.width_mult, 8)
            layers.append(InvertedResidual(input_channel, output_channel, k, s, t, use_se, use_hs))
            input_channel = output_channel
        self.features = nn.Sequential(*layers)
        # building last several layers
        self.conv = Conv2dBNHswish(input_channel, exp_size, kernel_size=1)
        self.fc_inp = exp_size
        self.initFc()
        initWeightsKaiming(self)

    def forward(self, x):
        x = self.features(x)
        x = self.conv(x)
        x = self.pool(x)
        x = x.view(x.size(0), -1)
        x = self.classifier(x)
        return x

    def initFc(self):
        self.pool = nn.AdaptiveAvgPool2d((1, 1))
        output_channel = makeDivisible(self.last_output_channel * self.width_mult, 8) if self.width_mult > 1.0 else self.last_output_channel
        self.classifier = nn.Sequential(
            nn.Linear(self.fc_inp, output_channel),
            hswish(),
            nn.Dropout(0.2),
            nn.Linear(output_channel, self.class_num),
        )

    def auditConfig(self):
        self.last_output_channel = 1024
        self.cfgs = [
            # k, t, c, SE, HS, s 
            [3,    1,  16, True, False, 2],
            [3,  4.5,  24, False, False, 2],
            [3, 3.67,  24, False, False, 1],
            [5,    4,  40, True, True, 2],
            [5,    6,  40, True, True, 1],
            [5,    6,  40, True, True, 1],
            [5,    3,  48, True, True, 1],
            [5,    3,  48, True, True, 1],
            [5,    6,  96, True, True, 2],
            [5,    6,  96, True, True, 1],
            [5,    6,  96, True, True, 1],
        ]


class MobileNetV3Large(MobileNetV3):
    def auditConfig(self):
        self.last_output_channel = 1280
        self.cfgs = [
            # k, t, c, SE, HS, s 
            [3,   1,  16, False, False, 1],
            [3,   4,  24, False, False, 2],
            [3,   3,  24, False, False, 1],
            [5,   3,  40, True, False, 2],
            [5,   3,  40, True, False, 1],
            [5,   3,  40, True, False, 1],
            [3,   6,  80, False, True, 2],
            [3, 2.5,  80, False, True, 1],
            [3, 2.3,  80, False, True, 1],
            [3, 2.3,  80, False, True, 1],
            [3,   6, 112, True, True, 1],
            [3,   6, 112, True, True, 1],
            [5,   6, 160, True, True, 2],
            [5,   6, 160, True, True, 1],
            [5,   6, 160, True, True, 1]
        ]

class MobileNetV3Ocr(MobileNetV3):
    def auditConfig(self):
        self.last_output_channel = 1024
        self.cfgs = [
            # k, t, c, SE, HS, s 
            [3,    1,  16, True, False, 1],
            [3,  4.5,  24, False, False, (2,1)],
            [3, 3.67,  24, False, False, 1],
            [5,    4,  40, True, True, (2,1)],
            [5,    6,  40, True, True, 1],
            [5,    6,  40, True, True, 1],
            [5,    3,  48, True, True, 1],
            [5,    3,  48, True, True, 1],
            [5,    6,  96, True, True, (2,1)],
            [5,    6,  96, True, True, 1],
            [5,    6,  96, True, True, 1],
        ]
    def initFc(self):
        self.pool = nn.MaxPool2d(2)

    def forward(self,x):
        x = self.features(x)
        x = self.conv(x)
        x = self.pool(x)

        b, c, h, w = x.size()
        assert h == 1, "the height of conv must be 1"

        x = x.squeeze(2)        # b *c * width
        x = x.permute(2, 0, 1)    # [w, b, c]
        return x



import base64
import hashlib
import os
import sys
import string
import random
from .log import LOG

try:
    from Crypto import Random
    from Crypto.Cipher import AES
except:
    LOG.logE("Crypto package missing, you must install it with [pip3 install pycrypto]", exit=True)

class SyszuxCipher(object):
    def __init__(self, key):
        self.aes_bs = AES.block_size
        LOG.logI('Using block size = {}'.format(self.aes_bs))
        self.key = hashlib.sha256(key.encode()).digest()
        LOG.logI('Hash of key={} is {}'.format(key, self.key))

    def encrypt(self, pt_byte):
        pt_byte = self._pad(pt_byte)
        iv = Random.new().read(AES.block_size)
        LOG.logI('iv: {}'.format(iv))
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return iv + cipher.encrypt(pt_byte)

    def decrypt(self, pt_byte):
        iv = pt_byte[:self.aes_bs]
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return self._unpad(cipher.decrypt(pt_byte[self.aes_bs:]))#.decode('utf-8')

    def _pad(self, s):
        return s + str.encode((self.aes_bs - len(s) % self.aes_bs) * chr(self.aes_bs - len(s) % self.aes_bs))

    def __call__(self, path, mode='encrypt'):
        filename, file_extension = os.path.splitext(path)

        if mode == 'encrypt':
            output_filename = "{}_syszux.so".format(filename)
        else:
            output_filename = "{}.{}".format(filename, 'pt')

        if not os.path.exists(path):
            LOG.logE('File does not exist at path: %s'.format(path), exit=True)

        pt_byte = None
        with open(path, 'rb') as f:
            pt_byte = f.read()

        if mode == 'encrypt':
            pt_byte = self.encrypt(pt_byte)
        else:
            pt_byte = self.decrypt(pt_byte)

        with open(output_filename, 'wb') as f:
            f.write(pt_byte)

        LOG.logI('Saved with key={} to {}'.format(self.key, output_filename))

    @staticmethod
    def _unpad(s):
        return s[:-ord(s[-1:])]

if __name__ == "__main__":
    modes = ['encrypt', 'decrypt']
    if len(sys.argv) < 3:
        LOG.logE('Usage: {} <file_path> <mode> <key>'.format(sys.argv[0]), exit=True)

    file = sys.argv[1]
    mode = sys.argv[2]
    if mode not in modes:
        LOG.logE("Unsupported mode: {}".format(mode), exit=True)
 
    key = sys.argv[3]
    sc = SyszuxCipher(key)

    LOG.logI("You will use {} to {} {}".format(key, mode, file))
    sc(file, mode)
